/*
 * XRel.java
 * project: XRel
 * @author Fabrizio Bisi
 * Created on 10-giu-03
 */
package xrel;

import java.util.HashSet;

import xrel.analyzer.ElementAutomata;
import xrel.analyzer.PatternMatching;
import xrel.analyzer.SemanticChecks;
import xrel.analyzer.TABuildPhase;
import xrel.analyzer.TADifference;
import xrel.analyzer.TAProduct;
import xrel.analyzer.TA;
import xrel.analyzer.TAUnion;
import xrel.analyzer.Translate;
import xrel.analyzer.WellFormedness;
import xrel.parser.ASTProgram;
import xrel.parser.ParseException;
import xrel.parser.SimpleNode;
import xrel.parser.SpecialNames;
import xrel.parser.SymElementKind;
import xrel.parser.SymTable;
import xrel.parser.XRelParser;
import xrel.xml.XmlDoc;

/**
 * The main class of the program.
 * @author Fabrizio Bisi
 */
public class XRel implements SpecialNames {
	static private boolean specialChecking = true; // enable diagnostics checks 
	private static final String EXENAME = "XRel";
	private static final String USAGE = 
		"Usage: java [OPTIONS] " +  EXENAME + 
    " <XRel file> [<XML document>]\n" +
    "If you omit the second argoment only the static part is run.\n\n" +
    "OPTIONS: \n" +
		"-v (--verbose): show information about phases of the program and " + 
			"internal intermediate results\n" + 
//    "-d (--debug):   show debug information (internal details)\n" +
//		"To fully know what the program does you must specify both.\n" + 
		"-o <filename> (--output): export output to XML format\n" +
		"-t (--showtimes): show a final table with global times\n" + 
		"Note that you can't write options together " +
		"(e.g. you can't write -vt or -tv but you need to write -v -t or -t -v)\n\n" +
		"For more information see the XRel home page " +
		"http://www.cs.unibo.it/~bisi/thesis.html";
	public static final String VERSION_NUMBER = "0.12";
	private static final String FULLNAME =
	  "XML Regular Expressions Language " + VERSION_NUMBER + ": ";
	private static final String HELPSTRING = FULLNAME + "\n" + USAGE;
	protected static final String DEFAULT_PARSER_NAME = "xmlParser.Xerces";
	// put here command-line arguments
	private static String 
		XMLFile, 
		SrcFile,
		exportXMLfile=null;
	// command-line options
	private static boolean 
		verboseMode, // -v
		requireTimes, // -t
//	debugMode, // -d
		debugBoolOps, //	-a A1 A2 (undocumented) test A1xA2 and A1\A2
		internalFormat, // --di  (undocumented) enable internal format in debug mode
		debugParseTree, // --dpt (undocumented) show parse tree
		debugSymtab,    // --dst (undocumented) show final symbol table 
		heavyDebugSymtab,  // --hdst (undocumented) see symtab modifies in details
		debugElementAutomata, // --da (undocumented) see final element automata
		debugSemChks,   // --dsc (undocumented) show semantic checks debug info
		debugRunTime;   // --drt (undocumented) show run-time debug info
	private static String 
	  automaton1,automaton2; // used with debugBoolOps
	private static boolean
	  staticDebugMode; // on if at least one of static debug options is on 
	
	public static void main(String args[]) {
		long beforeTime; // used to compute times
		long
			// static phase times
			xreLoadTime=0,					// load  .xre file
			xreParseTime=0, 				// parse .xre file
			syntaxChecksTime=0,			// syntax checks and sym. table creation
			globalSymtabTime=0, 		// global symtab processing time
			desugaringTime=0,				// remove ground types and desugaring
			normalizePatternsTime=0, // normalize patterns
			mergeVarsTime=0,				// merging variables
			buildAutomataTime=0,		// initial building
//  		globalAutomataTime=0,		// global automata time
//			epsAutomataTime=0,			// remove epsilon transitions
//			uselessAutomataTime=0,	// remove useless states and complete
			semanticChecksTime=0,		// semantic checks
			// run-time phase times
		  xmlLoadTime=0,					// load & parse .xml file
		  xmlTranslateTime=0,			// translate DOM tree in an internal tree
		  //xmlNormalizeTime=0,			// internalize format of the XML doc
			patMatchTime=0;					// pattern matching
		
		if (ParseArgs(args) != -1) {
			// some static debug option is on? 
			staticDebugMode = 
				debugParseTree | debugSymtab | heavyDebugSymtab | 
				debugElementAutomata | debugSemChks;
			
			/* -----------
			 * XRel parser
			 * -----------
			 */
			if (verboseMode || staticDebugMode) {
				System.out.println("<<<<");
				System.out.println(FULLNAME + "STATIC PHASE started\n");
			}

			// initialize the parser
			XRelParser pr = null;
			try {
				if (verboseMode) 
					System.out.print("Reading program " + SrcFile + "...");
				beforeTime = System.currentTimeMillis();
				pr = new XRelParser(new java.io.FileInputStream(SrcFile));
				xreLoadTime = System.currentTimeMillis() - beforeTime;
				if (verboseMode) 
					System.out.println("completed (" + xreLoadTime + " msec.)");
			}
			catch(java.io.FileNotFoundException fnfe) {
				XreError.showerror_and_exit("file not found");
			}
	        
			// initialize symbol table
			SymTable symtab = new SymTable();
			
			// parse tree
			ASTProgram rt = null; // root node
			try {
				if (verboseMode) 
					System.out.print("Building of the syntax tree...");
				// parse the input file
				beforeTime = System.currentTimeMillis();
				rt = pr.Program(symtab);
				xreParseTime = System.currentTimeMillis() - beforeTime;
				if (verboseMode) 
					System.out.println("completed (" + xreParseTime + " msec.)");
			}
			catch(ParseException pe) {
				System.err.println("encountered errors during parse");
				pe.printStackTrace();
				System.exit(1);
				return;
			}
			
			// Writes out the syntax tree
			if (debugParseTree) {
				System.out.println("\nHere's the syntax tree:");
				System.out.println("-----------------------\n");
				rt.dumpIntTree(System.out,"");
				System.out.println();
			}
			
			// Now perform some syntax checks
			if (verboseMode)
				System.out.print("Syntax checks...");
			beforeTime = System.currentTimeMillis();
			// add predefined types 
			Translate.addPredefined(symtab);
			// collect syntax errors
			symtab.checkErrors(rt); 
			// and check linearity of patterns
			WellFormedness.checkLinearity(symtab,rt); 
			// now check that type declarations are not top-level recursive 
			WellFormedness.disconnectedness(symtab,rt); 
			syntaxChecksTime = System.currentTimeMillis() - beforeTime;
			if (verboseMode)
				System.out.println("done (" + syntaxChecksTime + " msec.)");
			
			if (symtab.foundErrors()){
				if (verboseMode) {
					System.out.println("\nFound errors, here's the list:");
					System.out.println("------------------------------\n");
				}
				else
					System.out.println("\nListing errors:");
				symtab.showErrors(System.out,rt);
			}
			else {
				if (symtab.foundWarnings()){
					symtab.showErrors(System.out,rt);
				}

				// show the (original) symbol table
				if (heavyDebugSymtab) {
					System.out.println("\nNow we see the original symbol table:");
					System.out.println("-------------------------------------\n");
					symtab.dump(System.out,
						SymElementKind.S_TYPES | SymElementKind.S_VARIABLES |
						SymElementKind.S_PATTERNS | SymElementKind.S_EXPRESSIONS,true,
						internalFormat);
					System.out.println();
				}
				
				// processing symbol table
				globalSymtabTime = System.currentTimeMillis();
				/* clear the parent pointers of patterns & types
				 * (detach them from the syntax tree).
				 * I kill the original syntax tree here! */
				symtab.detachPatterns(); 
				/* transform expressions in the program (types, patterns,
				 * variable types and expressions): desugar and remove ground 
				 * types */
				if (verboseMode) {
					System.out.print("Removing ground types from symbol table " + 
													 "and desugaring ...");
				}
				beforeTime = System.currentTimeMillis();
				Translate.desugar(symtab);
				desugaringTime = System.currentTimeMillis() - beforeTime;
				if (verboseMode)
					System.out.println("done (" + desugaringTime + " msec.)");
				if (heavyDebugSymtab) {
					System.out.println("\nThe symbol table after desugaring:");
					System.out.println("----------------------------------\n");
					symtab.dump(System.out,
						SymElementKind.S_TYPES | SymElementKind.S_VARIABLES |
						SymElementKind.S_PATTERNS | SymElementKind.S_EXPRESSIONS,true,
						internalFormat);
					System.out.println();
				}
				if (specialChecking) Translate.diagnosticChecks(symtab,0); 
				
				/* Now the difficult things: normalize patterns.
				 * I keep separate this phase because here the size of the vector 
				 * of patterns can grow
				 */
				if (verboseMode)
					System.out.print("Normalizing types and patterns...");
				beforeTime = System.currentTimeMillis();
				Translate.normalization(symtab);
				normalizePatternsTime = System.currentTimeMillis() - beforeTime;
				if (verboseMode)
					System.out.println("done (" + normalizePatternsTime + " msec.)");
				if (heavyDebugSymtab) {
					System.out.println("\nThe symbol table after pattern normalization:");
					System.out.println("---------------------------------------------\n");
					symtab.dump(System.out,
						SymElementKind.S_TYPES | SymElementKind.S_VARIABLES |
						SymElementKind.S_PATTERNS | SymElementKind.S_EXPRESSIONS,true,
						internalFormat);
					System.out.println();
				}
				if (specialChecking) Translate.diagnosticChecks(symtab,1); 

				// now merge variables
				if (verboseMode)
					System.out.print("Merging variables...");
				beforeTime = System.currentTimeMillis();
				symtab.mergeVariables();
				mergeVarsTime = System.currentTimeMillis() - beforeTime;
				// processing symbol table completed
				globalSymtabTime = System.currentTimeMillis() - globalSymtabTime;
				if (verboseMode) {
					System.out.println("done (" + mergeVarsTime + " msec.)");
				}
				if (debugSymtab || heavyDebugSymtab) {
					System.out.println("\nNow we see the (final) symbol table:");
					System.out.println("------------------------------------\n");
					// and show the internal symbol table
					symtab.dump(System.out,
						SymElementKind.S_TYPES | SymElementKind.S_VARIABLES |
						SymElementKind.S_PATTERNS | SymElementKind.S_EXPRESSIONS,true,
						internalFormat);
					System.out.println();
				}
				if (specialChecking) Translate.diagnosticChecks(symtab,1); 
				
				// build all the automata
				if (verboseMode)
					System.out.print("Building automata..");
				beforeTime = System.currentTimeMillis();
				ElementAutomata.buildAll(symtab,debugElementAutomata);
				buildAutomataTime = System.currentTimeMillis() - beforeTime;
				if (verboseMode)
					System.out.println("done (" + buildAutomataTime + " msec.)");
				if (debugElementAutomata) {
					System.out.println("\nNow we see the built automata:");
					System.out.println("------------------------------\n");
					ElementAutomata.dumpAutomata(symtab,System.out);
					System.out.println();
				}

				/* Now perform semantic checks: 
				 * - exahustivity check
				 * - irredundancy check
				 * - exact output type <: declared output type
				 * - ambiguity check */
				boolean ambImport = false;
				boolean semChksResult = true; 

				if (verboseMode) System.out.print("Semantic checks...");
				beforeTime = System.currentTimeMillis();
				// import pattern is ambiguous?
				ambImport =
				 SemanticChecks.isAmbiguousImport(symtab,System.out,debugSemChks);
				if (!ambImport && symtab.originalPatterns > 1) {
					semChksResult =
						SemanticChecks.checkTypeSwitch(symtab,
															true,true,true,true,System.out,debugSemChks);
				}
				semanticChecksTime = System.currentTimeMillis() - beforeTime;
				if (verboseMode)
					System.out.println("done (" + semanticChecksTime + " msec.)");

				if (symtab.foundErrors() || symtab.foundWarnings()) {
					if (verboseMode) {
						System.out.println("\nFound errors/warnings, here's the list:");
						System.out.println("---------------------------------------\n");
					}
					symtab.showErrors(System.out,rt);
				} else if (verboseMode || debugSemChks) {
					System.out.println("\nNow show semantic check results:");
					System.out.println("--------------------------------\n");
					if (!ambImport) {
						System.out.println("Import pattern ok!");
						if (semChksResult && symtab.originalPatterns > 1) 
							System.out.println("Typeswitch ok!");
					}
				}

				if (debugSemChks) symtab.showThings(System.out);
			}
			
//			System.out.print("\nNow I write variables on file " + SymtabFile + 
//											 "...");
//			// export symbol table
//			symtab.export(SymtabFile);
//			System.out.println("completed");
			
			if (verboseMode || staticDebugMode) {
				System.out.println("\n" + FULLNAME + "STATIC PHASE terminated");
				System.out.println(">>>>");
			}
				
			/* ---------------
			 *  RUN-TIME PHASE
			 * ---------------
			 */
			if (symtab.foundErrors()) 
				System.out.println("\nFound errors, execution terminated");
			else if (XMLFile != null) {
				if (verboseMode || debugRunTime) {
					System.out.println("\n" + "<<<<");
					System.out.println(FULLNAME + "RUN-TIME PHASE started\n");
				}
	
				// load the XML document, build and parse the DOM tree 
				if (verboseMode)
					System.out.print("Reading XML file " + XMLFile + "...");
				beforeTime = System.currentTimeMillis();
				boolean xmlOk = XmlDoc.loadFrom(XMLFile);
				xmlLoadTime = System.currentTimeMillis() - beforeTime;
				if (verboseMode)
					System.out.println("completed (" + xmlLoadTime + " msec.)");

				if (!xmlOk)
					System.out.println("\nFound errors, execution terminated");
				else {
					// now translate the DOM tree to an internal tree
					if (verboseMode)
						System.out.print("Converting to an internal format...");
					beforeTime = System.currentTimeMillis();
					SimpleNode doc = XmlDoc.convert();
					Translate.desugarElement(symtab,doc); // remove ground types
					xmlTranslateTime = System.currentTimeMillis() - beforeTime;
					if (verboseMode)
						System.out.println("done (" + xmlTranslateTime + " msec.)");

					// PROCESSING: pattern matching 
					if (verboseMode)
						System.out.print("Now search matches...");
					beforeTime = System.currentTimeMillis();
					int matchingClause =
					 PatternMatching.matchWith(symtab, doc,System.out,debugRunTime);
					patMatchTime = System.currentTimeMillis() - beforeTime;
					if (verboseMode)
						System.out.println("done (" + patMatchTime + " msec.)");
					
					// Show the input file
					if (verboseMode || debugRunTime) {
						System.out.println("\nHere's the XML file:");
						System.out.println("--------------------\n");
						if (internalFormat)
							doc.dumpIntTree(System.out,""); // the internal form
						else {
							doc.writeExpression(System.out,"",symtab); // the normal form
							System.out.println();
						}
					}
					
					// finally show the results
					if (verboseMode || debugRunTime) {
						System.out.println("\nResult:");
						System.out.println("-------\n");
					}
					if (matchingClause != -1) {
						if (verboseMode) {
							System.out.println("The input document MATCHES " + 
									 "with the patterns of the program.");
							if (matchingClause > 0)
									System.out.println("Specifically clause " + 
																matchingClause + " matches.");
							System.out.println("\nList of the bound variables " + 
									 "in the program:");
							symtab.showBoundVariables(System.out,debugRunTime,internalFormat);
								if (matchingClause > 0)
							System.out.println("\nOutput:");
						}
						SimpleNode outValue = PatternMatching.getEvaluation();
						if (outValue != null) {
							outValue.writeExpression(System.out,"",symtab);
							System.out.println();
							if (exportXMLfile != null) {
								if (verboseMode)
									System.out.println("\nNow exporting XML format to " +
									                   exportXMLfile);
								outValue.exportExpression(exportXMLfile,symtab);
							}
						} 
						else System.out.println("<empty output>");
					}
					else
						System.out.println("The input document DOESN'T match with the " +
							"patterns of the program.");
				}
				
				if (verboseMode || debugRunTime) { 
					System.out.println("\n" + FULLNAME + "RUN-TIME PHASE terminated");
					System.out.println(">>>>");
				}
			}
			
			if (requireTimes) {
				System.out.println("\nTimes:");
				if (verboseMode) {
					System.out.println("------\n");
				}
				long staticTime = xreLoadTime + xreParseTime + syntaxChecksTime + 
				                  globalSymtabTime + buildAutomataTime + 
				                  semanticChecksTime;
				long dynamicTime = xmlLoadTime + xmlTranslateTime + patMatchTime;
				System.out.println("Total static time:  " + staticTime + " msecs");
				System.out.println("Total dynamic time: " + dynamicTime + " msecs");
				System.out.println("Total time:         " + 
				                   (staticTime + dynamicTime) + " msecs");
			}
			
			if (debugBoolOps) {
				System.out.println("\nAutomata debugging");	
				System.out.println("------------------\n");
				
				// automaton 1
				TA a1 = symtab.getSimpleAutomaton(automaton1,null,
																TABuildPhase.FULL_BUILD,true
																/*heavyDebugAutomata*/);
				if (a1 == null) {
					System.out.println("Automaton " + automaton1 + "not found!");
					System.exit(1);
				}
				else {
					System.out.println("Automaton " + automaton1 + ":");
					a1.dump(System.out);
					// now check if the automaton is empty
					System.out.println("\nNow check if automaton " + a1 + 
														 " is empty");
					if (a1.isNull())
						System.out.println("The recognized language is empty");
					else
						System.out.println("The recognized language is NOT empty");
				}
				
				// automaton 2
				TA a2 = symtab.getSimpleAutomaton(automaton2,null,
																TABuildPhase.FULL_BUILD,true
																/*heavyDebugAutomata*/);
				if (a2 == null) {
					System.out.println("Automaton " + automaton2 + "not found!");
					System.exit(1);
				}
				else if (!automaton1.equals(automaton2)) {
					System.out.println("\nAutomaton " + automaton2 + ":");
					a2.dump(System.out);
					// now check if the automaton is empty
					System.out.println("\nNow check if automaton " + a2 + 
														 " is empty");
					if (a2.isNull())
						System.out.println("The recognized language is empty");
					else
						System.out.println("The recognized language is NOT empty");
				}
				
				// union
				HashSet uNames = new HashSet();
				uNames.add(automaton1); uNames.add(automaton2);
				TAUnion taUnion = new TAUnion(uNames,symtab,true,System.out);
				System.out.println("\nUnion between " + 
				automaton1 + " and " + automaton2 + ":");
				taUnion.update(TABuildPhase.ALL_IN_ONE);
				taUnion.dump(System.out);
				if (taUnion.isNull())
					System.out.println("Union is empty");
				else
					System.out.println("Union is NOT empty");
							
				// product
				TAProduct taProduct =
					new TAProduct(a1,a2,symtab,true/*heavyDebugAutomata*/,System.out,true);
				taProduct.update(TABuildPhase.SIMPLE_BUILD);
				System.out.println("\nIntersection between " + 
				automaton1 + " and " + automaton2 + ":");
				taProduct.dump(System.out);
				// Finally remove useless states
				System.out.println("\nNow removing useless states:");
				taProduct.update(TABuildPhase.ALL_IN_ONE);
				taProduct.dump(System.out);
				
				System.out.println("\nNow check if intersection is empty");
				if (taProduct.isNull())
					System.out.println("Intersection is empty");
				else
					System.out.println("Intersection is NOT empty");
				
				// difference
				TADifference taDiff =
				 symtab.getDiffAutomaton(a1,a2,TABuildPhase.SIMPLE_BUILD,
				 												 true/*heavyDebugAutomata*/); 
				System.out.println("\nDifference between " + automaton1 + " and " + 
													 automaton2 + ":");
				taDiff.dump(System.out);
				// Finally remove useless states
				System.out.println("\nNow removing useless states:");
				taDiff.update(TABuildPhase.ALL_IN_ONE);
				taDiff.dump(System.out);
				if (taDiff.isNull())
					System.out.println("Difference is empty");
				else
					System.out.println("Difference is NOT empty");
				
				if (true/*heavyDebugAutomata*/) {
					System.out.println("\nHere's the final situation:");
					symtab.showThings(System.out);
				} 
			}
		}
		else {
			System.err.println("\n" + HELPSTRING);
			System.exit(1);
		}
	}
	
	// process command-line arguments
	private static int ParseArgs(String args[]) {
		final int 
		  minArgs = 1,
		  maxArgs = 2;
		int counter = 0;
		String localargs[] = new String[maxArgs];
		
		exportXMLfile = null;
//		debugMode = false;
		verboseMode = false;
		requireTimes = false;
		debugParseTree = false;
		debugSymtab = false;
		heavyDebugSymtab = false;
		debugElementAutomata = false;
		debugSemChks = false;
		debugRunTime = false;
		internalFormat = false;
		debugBoolOps = false;
		
		// collect real arguments and options and separate it
		for (int i = 0; i < args.length; i++) {
			if (args[i].charAt(0) == '-') {
				if ((args[i].equals("-v")) || (args[i].equals("--verbose"))) 
					verboseMode = true;
//				else if ((args[i].equals("-d")) || (args[i].equals("--debug"))) 
//					debugMode = true;
				else if ((args[i].equals("-t")) || (args[i].equals("--showtimes"))) 
					requireTimes = true;
				else if ((args[i].equals("-o")) || (args[i].equals("--output"))) {
					exportXMLfile = args[i+1];
					i++;
				}
				// UNDOCUMENTED DEBUG OPTIONS
				else if ((args[i].equals("-a")) || (args[i].equals("--automata"))) {
					// test product and difference algorithms with two patterns
					debugBoolOps = true;
					automaton1 = args[i+1];
					automaton2 = args[i+2];
					i += 2;
				}
				else if (args[i].equals("--dpt")) {
					// debug info on parse tree on
					debugParseTree = true;
				}
				else if (args[i].equals("--dst")) {
					// partial debug info on symbol table on
					debugSymtab = true;
				}
				else if (args[i].equals("--hdst")) {
					// full debug info on symbol table on
					heavyDebugSymtab = true;
				}
				else if (args[i].equals("--da")) {
					// debug info on automata on
					debugElementAutomata = true;
				}
				else if (args[i].equals("--dsc")) {
					// debug info on semantic checks on
					debugSemChks = true;
				}
				else if (args[i].equals("--drt")) {
					// debug info on run-time algorithm on
					debugRunTime = true;
				}
				else if (args[i].equals("--di")) {
					// showing the internal format of patterns in debug info
					internalFormat = true;
				}
				else System.err.println("Error! Undefined option: " + args[i]);
			}
			else if (counter < maxArgs) localargs[counter++] = args[i];
			else {
				counter++; // signal error
				break;
			}
		}
		
		// now fill global variables with real arguments
		if (counter >= minArgs && counter <= maxArgs) { 
			SrcFile = localargs[0];
			if (counter >= 2) XMLFile = localargs[1];
			else XMLFile = null;
		}
		else counter = -1;
		return counter;
	}
}

interface DebugMode {
	final static int 
		NONE = 0,
		NORMAL = 1,
		VERBOSE = 2;
}
