/*
 * TA.java
 * project: XRel
 * @author Fabrizio Bisi
 * Created on 24-set-03
 */
package xrel.analyzer;

import java.io.PrintStream;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import xrel.XreError;
import xrel.parser.SymTable;

/**
 * <p>Tree automata are just like normal string automata but elements are 
 * trees instead of characters. 
 * An automaton is a quadruple (Q,I,F,T) formed by a set Q of states, 
 * a subset I of Q that represents the set of the initial states, a 
 * subset F of Q that represents the set of final states and a set of 
 * transitions T. </p>
 * <p>Transitions can be either normal labeled transitions or epsilon 
 * (null) transitions. A normal labeled transition has the form:
 * <blockquote><pre>
 *   q1 = l[content] --> q2</pre>
 * </blockquote>  
 * and its meaning is: transit from state q1 to state q2 when the next 
 * input element is a tag with name "l" and with the specified content.
 * An epsilon transition has the form:
 * <blockquote><pre>
 *   q1 --> q2</pre>
 * </blockquote>
 * and its meaning is: non deterministically transit from state q1 to 
 * state q2. </p>
 * @author Fabrizio Bisi
 */
abstract public class TA implements Cloneable,TABuildPhase {
	// enable additional run-time checks
	static final private boolean securityCheck = true;
	// Building phases flags
	protected boolean 
		isBuilt=false; 			// the automaton is built
	private boolean 
		noEpsTrans=false,		// epsilon transitions removed
		noUnreach=false,    // unreachable states removed
		noUnmatch=false,		// unmatchable states removed
		isComplete=false;   // closure of the automaton completed
	private boolean
	  isNullF=false,			// is nullChecked=true this is the result
	  nullChecked=false;  // emptiness test has been invoked?
	// where to search other automata
	protected SymTable symtab;
	// debug fields
	protected boolean debug;
	protected PrintStream out;
	
	/**
	 * <p>A unique name for each automaton. 
	 * It's used as a unique key to identify the automaton. </p>
	 * <p>N.B.: two distinct automata with the same structure have distinct
	 * names. </p> 
	 */
	protected String name;
	
	/** all the states of the automaton. They're TAState objects */
	protected Vector states;
	/** the initial states of the automaton. Note that
	 * it's a subset of the states of the automaton.
	 */
	protected Vector initialStates;
	/** the final states of the automaton. Note that
	 * it's subset of the states of the automaton.
	 */
	protected Vector finalStates;
	/** epsilon or null transitions (this is a NFA). 
	 * They're TAEpsTransition objects.
	 */
	protected Vector epsilonTransitions;
	/** normal transitions of the NFA. They're TATransition objects.
	 * Note that the total number of transitions of an automaton is 
	 * given by the number of normal transitions plus the number of
	 * the epsilon transitions.
	 */
	protected Vector transitions;
	
	/**
	 * Initializes an empty tree automaton. Subclasses need to explicitly
	 * introduce one or more methods to build the automaton.
	 * Generally there is a method named build().
	 */
	public TA(String nm, SymTable st, boolean dbg, PrintStream psOut) {
		name = nm;
		states = new Vector();
		initialStates = new Vector();
		finalStates = new Vector();
		epsilonTransitions = new Vector();
		transitions = new Vector();
		debug = dbg;
		out = psOut;
		symtab = st;
		/* put on the list of automata only if this is not a subpattern
		 * (used by BuildTA of TASimple)
		 */
		if (!nm.equals("subpattern")) st.addAutomaton(this); 
		// clear flags?
	}
	
	public void setDebug(boolean dbg) {
		debug = dbg;
	}
	
	/**
	 * <p>Duplicates the structure of the current automaton.
	 * Specifically it clones states and (consequently) transitions so 
	 * that the resulting automaton will have different states. </p>
	 * <p>States need to be unique for each automaton as they have internal
	 * status information (the isFinal flag and the exiting transitions);
	 * to mantain coherent the automaton transitions need to be
	 * remapped to new states so also transitions need to be changed, 
	 * also if they could by itself to be shared between multiple automata 
	 * (they don't have status information).
	 * If the flag isComplete is true this function calls completeAutomtaton()
	 * to update status information of the states.
	 * @param A the automaton to duplicate
	 */
	public Object clone(String newN) {
		TA newAut;
		try {
			newAut = (TA)super.clone();
		}
		catch(CloneNotSupportedException e) {
			XreError.showbug_and_exit("absurd! Object.clone() doesn't exist?");
			return null;
		}
		newAut.name = newN;
		newAut.states = new Vector();
		newAut.initialStates = new Vector();
		newAut.finalStates = new Vector();
		newAut.transitions = new Vector();
		newAut.epsilonTransitions = new Vector();
		// map from old states to new states
		HashMap stateMap = new HashMap();
		// now clone states
		for (int i = 0; i < states.size(); i++) {
			TAState s = (TAState)states.elementAt(i);
			TAState newS = (TAState)s.clone();
			newAut.states.add(newS);
			if (initialStates.contains(s)) newAut.initialStates.add(newS);
			if (finalStates.contains(s))
				// I DON'T set the isFinal flag
				newAut.finalStates.add(newS); 
			stateMap.put(s,newS);
		}
		// convert epsilon transitions 
		for (int i = 0; i < epsilonTransitions.size(); i++) {
			TAEpsTransition t = (TAEpsTransition)epsilonTransitions.elementAt(i);
			TAState src = (TAState)stateMap.get(t.src);
			TAState dst = (TAState)stateMap.get(t.dst);
			TAEpsTransition newT = new TAEpsTransition(src,dst);
			newAut.epsilonTransitions.add(newT);
		}
		// now convert labeled transitions 
		for (int i = 0; i < transitions.size(); i++) {
			TATransition t = (TATransition)transitions.elementAt(i);
			TAState src = (TAState)stateMap.get(t.src);
			TAState dst = (TAState)stateMap.get(t.dst);
			// variables are referenced, not copied!
			TATransition newT = new TATransition(t.label,src,dst,t.content,t.variables);
			newAut.transitions.add(newT);
		}
		/* TODO problems with product and difference automata: they're built
		 * without isComplete = true but they're however complete
		 */
		if (newAut.isComplete) completeAutomaton();
		return newAut;
	}

	/**
	 * Returns the child automata, that is set of automata names specified in 
	 * transitions but itself.
	 * @return the set of child automata
	 */
	public HashSet getChildAutomata() {
		if (isBuilt) {
			HashSet subAutomata = new HashSet();
			for (int i=0; i< transitions.size(); i++) {
				TATransition t = (TATransition)transitions.elementAt(i);
				if (!t.content.equals(name)) subAutomata.add(t.content);
			}
			return subAutomata;
		} 
		else return null;
	}

	public HashSet getDescendantAutomata() {
		HashSet subAutomata = new HashSet();
		if (isBuilt) {
			/* include myself first! so that I never recur on myself
			 * (neither directly or indirectly through mutual rec.)
			 */
			subAutomata.add(name);
			// call the recursive function 
			addDescendantAutomata(subAutomata);
			/* now I can remove myself. I think it's better to remove
			 * myself because ambiguity checks need that check on myself
			 * comes before checks on true subautomata
			 */
			subAutomata.remove(name);
			return subAutomata;
		}
		else return null;
	}
	private void addDescendantAutomata(HashSet subAutomata) {
		if (isBuilt) {
			for (int i=0; i< transitions.size(); i++) {
				TATransition t = (TATransition)transitions.elementAt(i);
				if (!subAutomata.contains(t.content)) {
					subAutomata.add(t.content);
					TA sa = symtab.search_automaton(t.content);
					if (sa != null /* && sa.isBuilt */)
						sa.addDescendantAutomata(subAutomata);
				}
			}
		} 
	}
	
	/**
	 * <p>Subclasses must implement a matchTree function. </p> 
	 * <p>TODO this function will presumably change. </p>
	 * @param value the value to accept
	 * @return true if the value matches the automaton 
	 */
	//abstract public boolean matchTree(SimpleNode value);
	
	abstract public void build();
	
	protected String getName() {
		return name;
	}
	
	public String toString() {
		return name;
	}
	
	/**
	 * Two automata are equal if and only if they have the same internal
	 * name. This is because each automaton has a unique internal name.
	 * @param o the object to compare to
	 * @return true if we have the same automaton
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object o) {
		return (getClass() == o.getClass()) && getName().equals(((TA)o).getName());
	}

	/**
	 * Redefines the HashCode function of Java using the hashing code of
	 * the string that represents the internal name.
	 * I don't know if it's a good hashing functions but who cares?
	 * @return the hashing code of the automaton
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode() {
		return getName().hashCode();
	}
	
	// remove transitions with null subautomata
	public void simplify() {
		update(TABuildPhase.ALL_IN_ONE);
		
		if (isBuilt) {
			// Warning! the reverse order is needed! as I can delete from Vector
			for (int i = transitions.size() - 1; i >= 0; i--) {
				TATransition tr = (TATransition)transitions.elementAt(i);
				TA sa = symtab.search_automaton( tr.content);
				if (sa == null)
					XreError.showbug_and_exit("can't find automaton " +  tr.content);
				sa.update(TABuildPhase.ALL_IN_ONE);
				if (sa.isNull())
					transitions.remove(i);
			}
			noUnmatch = false; noUnreach = false; isComplete = false;
			update(TABuildPhase.ALL_IN_ONE);
		}
		else
			XreError.showbug_and_exit("can't simplify " +  name + " because is not built yet!");
	}
	
	// build the automata completely first!
	public boolean isNull() {
		if (!isComplete) {
			XreError.showbug_and_exit("attempt to test emptiness of " + 
																"an automaton (" + name + 
																") not yet completed");
			return false;
		}
		else if (nullChecked) return isNullF;
		else {
			// take the list of all the subautomata and include myself
			HashSet saList = getDescendantAutomata();
			if (debug) {
				out.println("Descendant automata of " + name+ ": " + saList);
			}
			saList.add(name); 
			/* In the fixpoint algorithm defined below the following mapping
			 * from automata names to boolean values will contain the partial 
			 * emptiness values of the automata.
			 */  
			HashMap map = new HashMap();
			/* Initialize mapping from automata to emptiness values in the 
			 * following way. Given a subautomaton sa:
			 * - if sa.nullChecked == true then set map[sa] = sa.isNullF
			 * - else if sa recognizes the empty string then set map[sa] = false
			 * - else map[sa] = true
			 * The algorithm would work well also filling all emptiness values
			 * with true (the inizialization above is an optimization)
			 */
			Iterator salIt = saList.iterator();
			while (salIt.hasNext()) {
				String saName = (String)salIt.next();
				TA sa = symtab.search_automaton(saName);
				if (sa == null) {
					XreError.showbug_and_exit("Automaton " + saName + " not found");
					return false;
				}
				if (sa.nullChecked) map.put(saName,new Boolean(sa.isNullF));
				else if (sa.recognizesEmptyString()) map.put(saName, Boolean.FALSE);
				else map.put(saName, Boolean.TRUE);
			}
			// now let we begin with the fixpoint algorithm
			boolean fixpoint; // we have reached the fixpoint?
			do {
				fixpoint = true;
				Iterator it = map.keySet().iterator();
				while (it.hasNext()) {
					String saName = (String)it.next();
					TA sa = symtab.search_automaton(saName);
					if (sa == null) {
						XreError.showbug_and_exit("Automaton " + saName + " not found");
						return false;
					}
					// fixpoint iteration
					if (((Boolean)map.get(saName)).booleanValue() != false &&
					    !sa.isNullYet(map)) {
						/* the empty value for sa is changed! update the map and
						 * force another loop */
						map.put(saName,Boolean.FALSE);
						fixpoint = false;
					} 
				}
			} while (!fixpoint);
			// updates the values of myself and of all the descendant automata 
			Iterator it = map.entrySet().iterator();
			while (it.hasNext()) {
				Map.Entry e = (Map.Entry)it.next();
				String saName = (String)e.getKey();
				TA sa = symtab.search_automaton(saName);
				if (sa == null) {
					XreError.showbug_and_exit("Automaton " + saName + " not found");
					return false;
				}
				boolean saIsNull = ((Boolean)e.getValue()).booleanValue();
				sa.isNullF = saIsNull;
				sa.nullChecked = true;
			}
			return isNullF;
		}
	}
	
	private boolean recognizesEmptyString() {
		for (int i=0; i<initialStates.size(); i++) {
			TAState s = (TAState)initialStates.elementAt(i);
			if (s.isFinal()) return true;
		}
		return false;
	}
	
	private boolean isNullYet(HashMap map) {
		/* the valid transitions are the transitions with
		 * a non-empty content automaton */  
		Vector validTransitions = new Vector();
		for (int i=0; i< transitions.size(); i++) {
			TATransition t = (TATransition)transitions.elementAt(i);
			if (!((Boolean)map.get(t.content)).booleanValue()) 
				validTransitions.add(t);
		}
		// validStates are states reachable from the the initial ones
		Vector validStates = new Vector();
		// add before all the initial states
		validStates.addAll(initialStates);
		// then add all the states reachable with valid transitions  
		boolean changed;
		do {
			changed = false;
			// Warning! the reverse order is needed! as I can delete from Vector
			for (int i = validTransitions.size() - 1; i >= 0; i--) {
				TATransition tr = (TATransition)validTransitions.elementAt(i);
				if (validStates.contains(tr.src)) {
					changed = true;
					validTransitions.remove(tr);
					// also add the destination state
					if(!validStates.contains(tr.dst)) validStates.add(tr.dst);
				}
			}
		} while(changed);
		// finally check if it exists at least a reachable final state
		for (int i = finalStates.size() - 1; i >= 0; i--) {
			TAState fs = (TAState)finalStates.elementAt(i);
			if (validStates.contains(fs)) return false; 
		}
		// I've not found any reachable final state, is null yet
		return true; 
	}
	
	public void update(int BuildPhases) {
		if ((BuildPhases & TABuildPhase.STEP_BUILD) > 0 && !isBuilt) {
			build();
			isBuilt = true;
		}
		if (isBuilt && (BuildPhases & TABuildPhase.STEP_NO_EPS_TRANS) > 0 &&
		    !noEpsTrans) {
			compute_closures(); 
			epsilon_elimination();
			noEpsTrans = true;
		}
		if (isBuilt && (BuildPhases & TABuildPhase.STEP_NO_UNREACH_STATES) > 0 &&
		    !noUnreach) {
			remove_unreachable_states();
			noUnreach = true;
		}
		if (isBuilt && (BuildPhases & TABuildPhase.STEP_NO_UNMATCHED_STATES) > 0 &&
		    !noUnmatch) {
			remove_unmatched();
			noUnmatch = true;
		}
		if (isBuilt && (BuildPhases & TABuildPhase.STEP_COMPLETE_AUT) > 0 &&
		    !isComplete) {
			completeAutomaton();
			isComplete = true;
		}
	}
	
	public void force(int BuildPhases) {
		if ((BuildPhases & TABuildPhase.STEP_BUILD) > 0)
			isBuilt = false;
		if ((BuildPhases & TABuildPhase.STEP_NO_EPS_TRANS) > 0)
			noEpsTrans = false;
		if ((BuildPhases & TABuildPhase.STEP_NO_UNREACH_STATES) > 0)
			noUnreach = false;
		if ((BuildPhases & TABuildPhase.STEP_NO_UNMATCHED_STATES) > 0)
			noUnmatch = false;
		if ((BuildPhases & TABuildPhase.STEP_COMPLETE_AUT) > 0)
			isComplete = false;
		update(BuildPhases);
	}

	/**
	 * <p>For each state of the automaton compute the epsilon closure and
	 * update the information in the state object. </p>
	 * <p>The epsilon closure of a state is the state itself plus all the
	 * states reachable with only epsilon transitions.</p>
	 * @see xrel.analyzer.TAState#set_epsClos(HashSet e_clos)
	 */
	protected void compute_closures() {
		// computes all the epsilon_closures
		for (int i = 0; i < states.size(); i++) {
			HashSet closSet = new HashSet();
			TAState q = (TAState)states.elementAt(i);
			epsilon_closure(closSet,q); // private subroutine
			q.set_epsClos(closSet); // now update q
		}
	}
	
	/**
	 * <p>Removes the unreachable states from the automaton. 
	 * Unreachable states are the states that can't be reached from the
	 * initial states.</p>
	 * <p><b>Algorithm:</b></p>
	 * <p><ol>
	 * <li>Assign the initial states to the set of reachable states</li>
	 * <li>Add to the set of reachable states all the states reachable in
	 * a step that are not yet in the set</li>
	 * <li>Repeat step 2 until there are no more states to add to the set</li>
	 * <li>Discard the remaining states and update the set of final states</li>
	 * </ol></p>
	 * <p>N.B.: 
	 * <ul>
	 * <li>call epsilon_elimination() before this function</li>
	 * <li>In Ambig.ps (page 13) Hosoya says that exists a well known
	 * linear-time algorithm for useless states elimination and cites
	 * TATA (remember that useless states are both the unreachable and
	 * the unmatchable states). I've looked at TATA but I've found a
	 * polynomial algorithm (RED at page 23) equivalent to that I'm using 
	 * here. I don't know what Hosoya refers to (this algorithn is
	 * certainly NOT linear)</li> 
	 * </ul></p>
	 */
	protected void remove_unreachable_states() {
		// add before all the initial states
		states.clear();
		states.addAll(initialStates);
		
		// then filter out transitions 
		Vector newTransitions = new Vector();
		boolean changed;
		do {
			changed = false;
			// Warning! the reverse order is needed! as I can delete from Vector
			for (int i = transitions.size() - 1; i >= 0; i--) {
				TATransition tr = (TATransition)transitions.elementAt(i);
				if (states.contains(tr.src)) {
					changed = true;
					newTransitions.add(tr);
					transitions.remove(tr);
					// also add the destination state
					if(!states.contains(tr.dst)) states.add(tr.dst);
				}
			}
		} while(changed);
		transitions = newTransitions;
		
		// finally update the final states
		for (int i = finalStates.size() - 1; i >= 0; i--) {
			TAState fs = (TAState)finalStates.elementAt(i);
			if (!states.contains(fs)) 
				finalStates.remove(i);
		}
		
		// update flag
		noUnreach = true;
	}
	
	/**
	 * <p>Eliminates the epsilon transitions from the automaton. </p>
	 * <p><b>Algorithm:</b></p>
	 * <p><ol>
	 * <li>Add to the set of final states each state that has a final 
	 * state in its epsilon closure</li>
	 * <li>Add to each state the exiting transitions that exit from all
	 * the states in its epsilon closure</li>
	 * </ol></p>
	 * <p>Note that this is the standard algorithm you can find also in 
	 * string automata.</p>
	 * <p>N.B.:
	 * <ul>
	 * <li>call compute_closures before this function or a program bug
	 * will be signaled</li>
	 * <li>Hosoya says that this operation doesn't preserve the ambiguity</li>
	 * </ul></p>
	 */
	protected void epsilon_elimination() {
		// expand final states
		for (int i = 0; i < states.size(); i++) {
			TAState q = (TAState)states.elementAt(i);
			HashSet eps_clos = q.get_epsClos();
			if (eps_clos == null) {
				XreError.showbug_and_exit("invoke compute_closures first!");
			}
			if (!finalStates.contains(q)) {
				Iterator itr = eps_clos.iterator();
				while (itr.hasNext()) {
					TAState p = (TAState)itr.next();
					if (finalStates.contains(p)) {
						finalStates.add(q);
						break;
					}
				}
			}
		}
		// expand transitions
		for (int i = transitions.size() - 1; i >= 0; i--) {
			TATransition t = (TATransition)transitions.elementAt(i);
			for (int j = 0; j < states.size(); j++) {
				TAState q = (TAState)states.elementAt(j);
				if (q != t.src && q.get_epsClos().contains(t.src)) {
					TATransition newt = 
						new TATransition(t.label,q,t.dst,t.content,t.variables);
					transitions.add(newt);
				}
			}
		}
		// remove epsilon transitions
		epsilonTransitions.clear();
		
		// update flag
		noEpsTrans = true;
	}
	
	/**
	 * For each state set the Transition set and the flag isFinal.
	 */
	protected void completeAutomaton() {
		for (int i = 0; i < states.size(); i++) {
			TAState s = (TAState)states.elementAt(i);
			// check if it's final
			boolean isFin = false;
			for (int j=0; j < finalStates.size(); j++) {
				if (s == finalStates.elementAt(j)) {
					isFin = true;
					break;
				}
			}
			s.setFinal(isFin);

//			// check if it's initial
//			boolean isInit = false;
//			for (int j=0; j < initialStates.size(); j++) {
//				if (s == initialStates.elementAt(j)) {
//					isInit = true;
//					break;
//				}
//			}
//			s.setInitial(isInit);

			// set the transition set
			s.getTransitions().clear();
			for (int j=0; j < transitions.size(); j++) {
				TATransition t = (TATransition)transitions.elementAt(j);
				if (t.src == s) s.getTransitions().add(t);
			}
		} 
		
		// update flag
		isComplete = true;
	}
	
	/**
	 * <p>Removes the unmatched states from the automaton. 
	 * Unmatched states are states that can't be matched by any value.</p>
	 * <p><b>Algorithm:</b></p>
	 * <p>The intuition is to proceed backward from final states and
	 * mark all visited states.
	 * Note that this is the reverted remove_unreachable algorithm 
	 * (that is remove_unreachable proceeds forward from initial
	 * states to final states, as remove_unmatched proceeds backward
	 * from final states to initial states)</p>
	 * <p>In details:
	 * <ol>
	 * <li>Assign the final states to the set of matchable states</li>
	 * <li>Add to the set of matchable states all the states that take in
	 * a matchable state in a step and that are not yet in the set</li>
	 * <li>Repeat step 2 until there are no more states to add to the set</li>
	 * <li>Discard the remaining states and update the set of initial states</li>
	 * </ol></p>
	 */
	public void remove_unmatched() {
		// add before all the final states
		states.clear();
		states.addAll(finalStates);
		
		// then filter out transitions 
		Vector newTransitions = new Vector();
		boolean changed;
		do {
			changed = false;
			// Warning! the reverse order is needed! as I can delete from Vector
			for (int i = transitions.size() - 1; i >= 0; i--) {
				TATransition tr = (TATransition)transitions.elementAt(i);
				if (states.contains(tr.dst)) {
					changed = true;
					newTransitions.add(tr);
					transitions.remove(tr);
					// also add the source state
					if(!states.contains(tr.src)) states.add(tr.src);
				}
			}
		} while(changed);
		transitions = newTransitions;
		
		// finally update the initial states
		for (int i = initialStates.size() - 1; i >= 0; i--) {
			TAState s = (TAState)initialStates.elementAt(i);
			if (!states.contains(s)) initialStates.remove(i);
		}
		
		// update flag
		noUnmatch = true;
	}

	/**
	 * <p>Writes out every information about the automaton. </p>
	 * <p>Specifically:
	 * <ul>
	 * <li>States</li>
	 * <li>Initial states</li>
	 * <li>Final states</li>
	 * <li>Labeled transitions</li>
	 * <li>Epsilon transitions, if any</li>
	 * <li>Statistics (number of states and number of transitions)</li>
	 * </ul></p>
	 * @param out the stream to which print out
	 */
	public void dump(PrintStream out) {
		out.println("Name: " + getName());
		dumpStates(out,"States",states);
		dumpStates(out,"Initial states",initialStates);
		dumpStates(out,"Final states",finalStates);
		out.println("Transitions:");
		for (int i = 0; i < transitions.size(); i++) {
			out.println("  " + (TATransition)transitions.elementAt(i));
		}
		if (epsilonTransitions.size() > 0) {
			out.println("Null transitions:");
			for (int i = 0; i < epsilonTransitions.size(); i++) {
				out.println("  " + (TAEpsTransition)epsilonTransitions.elementAt(i));
			}
		}
//		out.println("Epsilon closures:");
//		for (int i = 0; i < states.size(); i++) {
//			TA_State q = (TA_State)states.elementAt(i);
//			System.out.println(q + ": " + q.get_epsClos());
//		}
//		out.println("State transitions:");
//		for (int i = 0; i < states.size(); i++) {
//			State s = (State)states.elementAt(i);
//			out.print("  ");
//			for (int j = 0; j < s.transitions.size(); j++) {
//				out.print((Transition)s.transitions.elementAt(j) + ", ");
//			}
//			out.println();
//		}
		// subautomata
		out.println("Child automata: " + getChildAutomata());
		// write out statistics (states+trans)
		out.println(getDebugInfo()); 
	}
	
	/* Prints out the states from the vector s to the stream out.
	 * Subroutine of dump().
	 */
	private void dumpStates(PrintStream out, String prefix, Vector s) {
		out.print(prefix + ": { " );
		for (int i = 0; i < s.size(); i++) {
			out.print((TAState)s.elementAt(i) + " ");
		}
		out.println("}");		
	}
	
	/* Gets the statistics of the automaton in a string.
	 * Subroutine of dump().
	 */
	private String getDebugInfo() {
		String rslt =
		  states.size() + " states and " + 
		  (transitions.size() + epsilonTransitions.size()) + " transitions";
		if (epsilonTransitions.size()>0)
			rslt += " (whose " + epsilonTransitions.size() + " are null)";
		return rslt;
	}
	
	/* 
	 * Adds a state to the automaton.
	 * Flags = INITIAL | FINAL
	 */
	protected void addState(TAState Q, int Flags) {
		states.add(Q);
		if ((Flags & TAPiece.INITIAL) > 0) {
			initialStates.add(Q);
			//Q.setInitial(true);
		}
		if ((Flags & TAPiece.FINAL) > 0) {
			finalStates.add(Q);
			//Q.setFinal(true);
		}
	}
	
	/*
	 * Adds a label transition from Q1 to Q2 with content "content" and label 
	 * "label" (X are the accumulators).
	 * Q1 and Q2 need to be already present in the automaton (and the content?)
	 */
	protected void addTrans(TAState Q1, TAState Q2, String content, 
															String label, HashSet X) {
		if (securityCheck) {
			if (! searchState(Q1)) {
				XreError.showbug_and_exit("cannot add a transiton " + 
								   "from a non-existent state");
			}
			else if (!searchState(Q2)) {
				XreError.showbug_and_exit("cannot add a transiton " + 
								   "to a non-existent state");
			}
			// TODO check also the content?
		}
		
		TATransition newTrans = new TATransition(label,Q1,Q2,content,X);
		transitions.add(newTrans);
	}
	
	/** Adds an epsilon transition from Q1 to Q2 */
	protected void addEpsTransition(TAState Q1, TAState Q2) {
		if (securityCheck) {
			if (! searchState(Q1)) {
				XreError.showbug_and_exit("cannot add a transiton " + 
								   "from a non-existent state");
			}
			else if (!searchState(Q2)) {
				XreError.showbug_and_exit("cannot add a transiton " + 
								   "to a non-existent state");
			}
		}
		
		TAEpsTransition newTrans = new TAEpsTransition(Q1,Q2);
		epsilonTransitions.add(newTrans);
	}
	
	/**
	 * Adds epsilon transitions from every state in V1 to every
	 * state in V2 
	 */
	protected void addEpsTransitions(Vector V1, Vector V2) {
		for (int i = 0; i < V1.size(); i++)
			for (int j = 0; j < V2.size(); j++) {
				TAEpsTransition newTrans = new 
					TAEpsTransition((TAState)V1.elementAt(i),(TAState)V2.elementAt(j));
				epsilonTransitions.add(newTrans);
			}
	}
	
	/**
	 * Adds epsilon transitions from every state in V1 to
	 * the state Q 
	 */
	protected void addEpsTransitions(Vector V1, TAState Q) {
		for (int i = 0; i < V1.size(); i++) {
			TAEpsTransition newTrans = new 
				TAEpsTransition((TAState)V1.elementAt(i),Q);
			epsilonTransitions.add(newTrans);
			}
	}
	
	/**
	 * Adds epsilon transitions from state Q to
	 * every state in V2 
	 */
	protected void addEpsTransitions(TAState Q, Vector V2) {
		for (int j = 0; j < V2.size(); j++) {
			TAEpsTransition newTrans = new 
				TAEpsTransition(Q,(TAState)V2.elementAt(j));
			epsilonTransitions.add(newTrans);
		}
	}
	
	/**
	 * <p>Merges the current automaton with another one. This function
	 * must be used only internally to build a bigger automaton from two
	 * subpieces that will be thrown away. To merge two complete automata
	 * that you want use indipendently use the function union. </p>
	 * <p>Note that this function let you copy only a part of the source 
	 * automaton with the parameter Flags.
	 * To copy the whole automaton use
	 * <blockquote>
	 * TAPiece.STATES | TAPiece.INITIAL | TAPiece.FINAL | TAPiece.TRANS
	 * </blockquote></p>
	 * 
	 * @param A the automaton to merge with the current one
	 * @param Flags the parts of A to copy (see above)
	 */
	protected void merge(TA A, int Flags) {
		if ((Flags & TAPiece.STATES) > 0) 
			states.addAll(A.states);
		if ((Flags & TAPiece.INITIAL) > 0) 
			initialStates.addAll(A.initialStates);
		if ((Flags & TAPiece.FINAL) > 0) 
			finalStates.addAll(A.finalStates);
		if ((Flags & TAPiece.TRANS) > 0) {
			transitions.addAll(A.transitions);
			epsilonTransitions.addAll(A.epsilonTransitions);
		}
	}
	
	/* search a state in an automaton.
	 * Used by addTrans & addEpsTransition for security checks!
	 */
	private boolean searchState(TAState Q1) {
		boolean found = false;
		TAState Q = null;

		for (int i = 0; i < states.size(); i++) {
			Q = (TAState)states.elementAt(i);
			/* TODO to search a state in an automaton 
			 * I perform a linear search comparing the pointers..
			 * Using hashing techniques instead?
			 */
			if (Q1 == Q) {
				found = true;
				break;
			}
		}
		return found;
	}
	
	/* Computes the epsilon closure of a state.
	 * Subroutine of compute_closures() 
	 * - see it for a description of the algorithm
	 */
	private void epsilon_closure(HashSet closure, TAState q) {
		closure.add(q);
		for (int i = 0; i < epsilonTransitions.size(); i++) {
			TAEpsTransition t = (TAEpsTransition)epsilonTransitions.elementAt(i);
			if (t.src == q && !closure.contains(t.dst)) 
				epsilon_closure(closure,t.dst);
		} 
	}
}