/*
 * TAProduct.java
 * project: XRel
 * @author Fabrizio Bisi
 * Created on 23-set-03
 */
package xrel.analyzer;

import java.io.PrintStream;

import xrel.XreError;
import xrel.parser.SpecialNames;
import xrel.parser.SymTable;

/**
 * TODO insert comment here
 * @author Fabrizio Bisi
 */
public class TAProduct extends TA implements SpecialNames {
	private TA firstAutomaton,secondAutomaton;
	/* this flag tells if to build self-product automata (AxA).
	 * Explicit building of self-product automata is useful in ambiguity
	 * checks.
	 * Note: there are risks in this use of forceSelfProduct?
	 * the value of the flag is propagated when new subautomata are built:
	 * for example if an old ... TODO complete or correct
	 */
	private boolean forceSelfProduct;
	
	public static String getProductName(TA A1, TA A2) {
		return A1.getName() + "x" + A2.getName();
	}
	public static String getProductName(String s1, String s2) {
		return s1 + "x" + s2;
	}
	
	public TAProduct(TA A1, TA A2, SymTable st, boolean dbg, PrintStream psOut, 
	                 boolean force) {
		super(getProductName(A1,A2),st,dbg,psOut);
		firstAutomaton = A1;
		secondAutomaton = A2;
		forceSelfProduct = force;
	}
	
	public void build() {
		// Clean states
		states.clear();
		initialStates.clear();
		finalStates.clear();
		transitions.clear();
		epsilonTransitions.clear();
		// First add initial states I1xI2
		for (int i=0; i<firstAutomaton.initialStates.size(); i++) {
			TAState s1 = (TAState)firstAutomaton.initialStates.elementAt(i);
			for (int j=0; j<secondAutomaton.initialStates.size(); j++) {
				TAState s2 = (TAState)secondAutomaton.initialStates.elementAt(j);
				TAProductState s = new TAProductState(s1,s2);
				states.add(s);
				initialStates.add(s);
//				s.setInitial(true);
				// final only if both states are final
				if (s1.isFinal() && s2.isFinal()) {
					s.setFinal(true);
					finalStates.add(s);
				}
			}
		}
		// Then iterate adding transitions and the other states
		for (int i=0; i<states.size(); i++) {
			TAProductState s = (TAProductState)states.elementAt(i);
			processState(firstAutomaton,secondAutomaton,s);
		}
	}
	
	/* Given the two original automata A1, A2 and a state alreay inserted
	 * in the product automaton insert all the other transitions that exit
	 * from the given state together with the destination states.
	 */
	private void processState(TA A1, TA A2,TAProductState s) {
		for (int i=0; i< s.getState1().getTransitions().size(); i++) {
			TATransition t1 =
			  (TATransition)s.getState1().getTransitions().elementAt(i);
			for (int j=0; j< s.getState2().getTransitions().size(); j++) {
				TATransition t2 =
				  (TATransition)s.getState2().getTransitions().elementAt(j);
				// the label of t1,t2 matches?
				String label = matchTransitionLabel(t1,t2);
				if (label != null) {
					/* the name of the automaton that matches the content is
					 * first x second (e.g. "p1xp2") */
					String contentName = getProductName(t1.content,t2.content);
					TA contentAutomaton =
					 getContentAutomaton(t1.content,t2.content);
					// create the new transition and possibly the destination state
					// the destination state already exists?
					TAProductState destS = getState(t1.dst,t2.dst);
					if (destS == null) {
						// create the destination state
						destS = new TAProductState(t1.dst,t2.dst);
						states.add(destS);
						// final only if both states are final
						if (t1.dst.isFinal() && t2.dst.isFinal()) {
							destS.setFinal(true);
							finalStates.add(destS);
						}
					}
					/* About the bound variables: it seems that Hosoya intersecate
					 * always types so that variable bindings are not important.
					 * The only time it intersecates a type and a pattern is
					 * in type inference where it intersecate a type for a patterns;
					 * for this I take bindings of the second one (it should be the
					 * pattern).  
					 */
					TATransition t = new TATransition(label,s,destS,contentName,t2.variables);
					transitions.add(t);
					s.getTransitions().add(t);
				}
			}
		}
	}
	
	private TA getContentAutomaton(String A1name, String A2name) {
		if (A1name.equals(A2name) && 
		    (!forceSelfProduct || name.equals(getProductName(A1name,A1name)))) {
			// now search the automaton
			TA contentAutomaton =
				symtab.search_automaton(A1name);
			contentAutomaton.update(TABuildPhase.STEP_NO_UNMATCHED_STATES);
			return contentAutomaton;
		}
		else {
			/* the name of the automaton that matches the content is
			 * first x second (e.g. "p1xp2") */
			String contentName = getProductName(A1name,A2name);
			// now search the automaton
			TA contentAutomaton =
				symtab.search_automaton(contentName);
			if (contentAutomaton == null) {
				// if it doesn't exist create it
				TA tA1 = symtab.search_automaton(A1name);
				TA tA2 = symtab.search_automaton(A2name);
				if (tA1 == null)
					XreError.showerror_and_exit("can't find automaton " + A1name);
				if (tA2 == null)
					XreError.showerror_and_exit("can't find automaton " + A2name);
				// create the product automaton
				contentAutomaton =
					new TAProduct(tA1,tA2,symtab,debug,out,forceSelfProduct);
				contentAutomaton.update(
					TABuildPhase.SIMPLE_BUILD | 
					// remove unmatched states (needful for ambiguity check) 
					TABuildPhase.STEP_NO_UNMATCHED_STATES);
				if (debug) {
					out.println("\nSubautomaton " + contentName + ":");
					contentAutomaton.dump(out);
				}
			}
			else
				// however be sure that the unmatched states are removed
				contentAutomaton.update(TABuildPhase.STEP_NO_UNMATCHED_STATES);

			return contentAutomaton;
		}
	}
	
	/* search a state and returns the state or null if not found */
	protected TAProductState getState(TAState s1, TAState s2) {
		// Warning! Note that I should use hashing techniques here
		TAProductState rslt = null;
		for (int i=0; i<states.size(); i++) {
			TAProductState s = (TAProductState)states.elementAt(i);
			if (s.getState1() == s1 && s.getState2() == s2) {
				rslt = s;
				break;
			}
		}
		return rslt;
	}

	/* Checks if two transitions match, i.e. if the label is the same.
	 * Contents can be different. Note that this function takes care to
	 * match the special "~" with all the tags but strings and string 
	 * constants with the type String.
	 * If transitions don't match this function returns null, otherwise 
	 * if transitions match it returns the common label (if one of two 
	 * labels is an instance and the other is a special symbol it returns 
	 * the instance). 
	 */
	private String matchTransitionLabel(TATransition t1, TATransition t2) {
		// N.B.: use matchesLabel of TA?
		String[] names = { t1.label, t2.label	};
		boolean[] isTag = {false, false };
		boolean[] isJolly = {false, false };
		
		for (int i=0; i<2; i++) {
			if (names[i].equals(AnyTagName)) { isTag[i] = true; isJolly[i] = true; }
			else if (names[i].startsWith(StrLiteralPrefix) && names[i].endsWith(StrLiteralSuffix)) {
				isTag[i] = false; isJolly[i] = false;
			}
			else if (names[i].equals(StringTypeName)) {
				isTag[i] = false; isJolly[i] = true;
			}
			else { isTag[i] = true; isJolly[i] = false; }
		}
		if (isTag[0] ^ isTag[1]) return null; // ^ = XOR (are different)
		else if (isJolly[0]) return names[1];
		else if (isJolly[1]) return names[0];
		else if (names[0].equals(names[1])) return names[0];
		else return null;
	}
}