/*
 * TASimple.java
 * project: XRel
 * @author Fabrizio Bisi
 * Created on 24-set-03
 */
package xrel.analyzer;

import java.io.PrintStream;
import java.util.HashSet;

import xrel.XreError;
import xrel.parser.ASTAs;
import xrel.parser.ASTComma;
import xrel.parser.ASTEmpty;
import xrel.parser.ASTIdentifier;
import xrel.parser.ASTOperator;
import xrel.parser.ASTPar;
import xrel.parser.ASTTag;
import xrel.parser.SimpleNode;
import xrel.parser.SpecialNames;
import xrel.parser.SymElement;
import xrel.parser.SymElementKind;
import xrel.parser.SymTable;

/**
 * Tree automata built from this class are either based upon an element
 * of the symbol table of the program (that can be a pattern, a type, 
 * a variable or an expression) or the union of a set of these elements. 
 * The name distingush them from more exotic automata like product and
 * difference automata.
 * @author Fabrizio Bisi
 */
public class TASimple extends TA implements SpecialNames {
	SymElement stElement=null;
	
	public TASimple(String nm, SymTable st, boolean dbg, PrintStream psOut, 
									SymElement el) {
		super(nm,st,dbg,psOut);
		stElement = el;
	}
	
	/**
	 * Builds the automaton for a symbol table element.
	 * @param el the symbol element of the pattern.
	 */
	public void build() {
		if (stElement != null) {
			SimpleNode Pi = stElement.getNode();
			HashSet X = new HashSet(); // empty set
			/* TODO before to build try to get if stEl != null
			 * (it's needed for unions!)
			 */
			TASimple A1 = buildTA(Pi,X);
			merge(A1,TAPiece.STATES | TAPiece.INITIAL | TAPiece.FINAL | TAPiece.TRANS);
		}
		else 
			XreError.showbug_and_exit("can't build this automaton, pointer to the element found null");
	}

	/**
	 * <p>This function checks if the automaton accepts a given value. </p>
	 * <p><b>Algorithm:</b></p>
	 * <p>matchTree fa solo il livello alto: confronta ogni stato iniziale
	 * dell'automa col primo tag del valore, chiamando la matchSequence.
	 * La matchSequence chiama se stessa ricorsivamente per andare avanti.
	 * </p>
	 * @param value the value to accept
	 * @param the debug status for the automaton
	 * @return true if the value matches against the automaton
	 */
	public boolean matchTree(SimpleNode value, boolean runtimeDebug) {
		debug = runtimeDebug;
		boolean found = false;
		for (int i = 0; i < initialStates.size(); i++) {
			TASimpleState current = (TASimpleState)initialStates.elementAt(i);
			found = matchSequence(firstTag(value),current);
			if (found) break;
		}
		return found;
	}
	
	/* Recursively checks the rest of the sequence.
	 * Checks the tags inside the value against the transitions. 
	 */
	private boolean matchSequence(SimpleNode value, TAState current) {
		if (value == null) {
			/* base case, the value is terminated, return true if we are in a 
			 * final state
			 */
			return current.isFinal();
		}
		else if (value.getClass() == ASTTag.class) {
			boolean found = false;
			for (int i = 0; i < current.getTransitions().size(); i++) {
				TATransition t = (TATransition)current.getTransitions().elementAt(i);
				String label = ((ASTTag)value).getName();
				// I will call matchTree with the first tag of the sequence
				SimpleNode contentNode = (SimpleNode)value.jjtGetChild(0);
				if (matchesLbl(label,t.label)) {
					if (debug) out.println("???????? attempting " + t.src + 
																				" --> " + t.dst);
					TASimple A =
					 (TASimple)symtab.search_automaton((String)t.content);
					if (A != null) {
						if (A.matchTree(contentNode,debug)) {
							found = matchSequence(nextTag(value),t.dst);
							if (found) {
								if (debug) out.println("++++++++ success    " + t.src + 
																							" --> " + t.dst);
								value.annotateValue(t.variables);
								break;
							}
							else {
								if (debug) out.println("-------- failure of " + t.src + 
																							" --> " + t.dst);
								value.eraseSubtreeAnnotations();
							} 
						}
					}
					else
						XreError.showbug_and_exit("can't find the automaton " + t.content);
				}
			}
			return found;
		}
		else {
			XreError.showbug_and_exit("malformed automaton");
			return false;
		}
	}

	static private boolean matchesLbl(String val, String pat) {
		if (pat.equals(AnyTagName)) {
			// "~" matches with any tag but not with strings
			if (val.startsWith(StrLiteralPrefix) && val.endsWith(StrLiteralSuffix)) return false;
			else return true;
		}
		if (pat.equals(StringTypeName)) {
			if (val.startsWith(StrLiteralPrefix) && val.endsWith(StrLiteralSuffix)) return true;
			else return false;
		}
		else return val.equals(pat);
	}
	
	/**
	 * <p>Returns the first tag in the value or null if it doesn't exist. </p>
	 * <p>Specifically:
	 * <ol>
	 * <li>returns the first child when value is a comma (in values commas 
	 * are not nested)</li>
	 * <li>returns the value when it's tag</li>
	 * <li>returns null when value is empty</li>
	 * </ol></p>
	 * @param value a tree value 
	 * @return the first tag of the value or null
	 */
	static private SimpleNode firstTag(SimpleNode value) {
		if (value.getClass() == ASTComma.class) {
			if (value.jjtGetNumChildren() > 0) {
				/* Warning! this code supposes that comma nodes have only 
				 * tags as children (no nested commas) - alternative: call
				 * return firstTag(value.jjtGetChild(0))
				 */
				return (SimpleNode)value.jjtGetChild(0);
			}
			else {
				XreError.showbug_and_exit("value tree is malformed");
				return null;
			}
		}
		else if (value.getClass() == ASTEmpty.class) {
			return null;
		}
		else if (value.getClass() == ASTTag.class) {
			return value;
		}
		else {
			XreError.showbug_and_exit("value tree has unknown node types");
			return null;
		}
	}
	
	/**
	 * <p>Returns the next tag in a tree value. We suppose this function is 
	 * called only in two circumnstances:
	 * <ol>
	 * <li>when we are in a sequence inside the tree value this function is
	 * called on the current tag of the sequence</li>
	 * <li>at the bottom of a sequence inside the tree value this function
	 * is called on a empty tag</li>
	 * </ol></p>
	 * @param value the current node in a tree value
	 * @return the next tag node or null
	 */
	static private SimpleNode nextTag(SimpleNode value) {
		if (value.getClass() == ASTComma.class) {
			XreError.showbug_and_exit("there is a bug in the " + 
                                "pattern matching algorithm");
			return null;
		}
		else if (value.getClass() == ASTEmpty.class) {
			return null;
		}
		else if (value.getClass() == ASTTag.class) {
			SimpleNode parent = (SimpleNode)value.jjtGetParent();
			if (parent == null) return null;
			else if (parent.getClass() == ASTComma.class) {
				SimpleNode rslt = null;
				for (int i=0; i < parent.jjtGetNumChildren(); i++) {
					if (value == parent.jjtGetChild(i)) {
						if (i < parent.jjtGetNumChildren()-1)
							rslt = (SimpleNode)parent.jjtGetChild(i+1);
						else rslt = null;
						break;
					}
				}
				return rslt;
			}
			else if (parent.getClass() == ASTTag.class) {
				/* this node is the unique child of another tag, so it doesn't
				 * another tag in the sequence
				 */
				return null;
			}
			else {
				XreError.showbug_and_exit("value tree has unknown node types"); 
				return null;
			}
		}
		else {
			XreError.showbug_and_exit("value tree has unknown node types");
			return null;
		}
	}
	
	private TASimple buildTA(SimpleNode node, HashSet PX) {
		TASimple A = new TASimple("subpattern",symtab,debug,out,null);
		HashSet X = new HashSet();
		X.addAll(PX);
		Class cl = node.getClass();
		
		if (cl == ASTEmpty.class) {
			TAState Q1 = new TASimpleState();
			A.addState(Q1,TAPiece.INITIAL | TAPiece.FINAL);
		}
	  else if (cl == ASTTag.class) {
			if (node.jjtGetNumChildren() == 1) {
				String label = ((ASTTag)node).getName();
				String identifier = ((ASTIdentifier)node.jjtGetChild(0)).getName();
				SymElement el =
				 symtab.search(SymElementKind.S_PATTERNS | SymElementKind.S_TYPES  
				 								| SymElementKind.S_EXPRESSIONS, identifier);
				if (el == null) {
					XreError.showbug_and_exit("can't find " + node.jjtGetChild(0));
					return null;
				}
				String automatonName = el.getAutomatonName();
				HashSet EmptySet = new HashSet();
				TAState Q1 = new TASimpleState();
				TAState Q2 = new TASimpleState();
				A.addState(Q1,TAPiece.INITIAL);
				A.addState(Q2,TAPiece.FINAL);
				A.addTrans(Q1,Q2,automatonName,label,X);
			}
			else {
				XreError.showbug_and_exit("tag with a wrong number of children");
				return null;
			}
		}
		else if (cl == ASTComma.class) {
			if (node.jjtGetNumChildren() < 2) {
				XreError.showbug_and_exit("comma with less of 2 children: how to handle it?");
			}
			else  {
				SimpleNode P1 = (SimpleNode)node.jjtGetChild(0);
				TASimple A1 = buildTA(P1,X);
				TASimple A2;
				A.merge(A1, TAPiece.STATES | TAPiece.INITIAL | TAPiece.TRANS);
				for (int i = 1; i < node.jjtGetNumChildren(); i++) {
					SimpleNode P2 = (SimpleNode)node.jjtGetChild(i);
					A2 = buildTA(P2,X);
					int Flags = TAPiece.STATES | TAPiece.TRANS;
					if (i == node.jjtGetNumChildren()-1) Flags |= TAPiece.FINAL;
					A.merge(A2, Flags);
					A.addEpsTransitions(A1.finalStates,A2.initialStates);
					A1 = A2;
				}
			}
		} 
		else if (cl == ASTPar.class) {
			for (int i = 0; i < node.jjtGetNumChildren(); i++) {
				SimpleNode P1 = (SimpleNode)node.jjtGetChild(i);
				TASimple A1 = buildTA(P1,X);
				A.merge(A1, 
					TAPiece.STATES | TAPiece.INITIAL | TAPiece.FINAL | TAPiece.TRANS);
			}
		}
		else if (cl == ASTOperator.class) {
			if (node.jjtGetNumChildren() == 1) {
				SimpleNode P1 = (SimpleNode)node.jjtGetChild(0);
				TASimple A1 = buildTA(P1,X);
				A.merge(A1, TAPiece.STATES | TAPiece.TRANS);
				TAState Q1 = new TASimpleState();
				TAState Q2 = new TASimpleState();
				A.addState(Q1,TAPiece.INITIAL);
				A.addState(Q2,TAPiece.FINAL);
				A.addEpsTransitions(Q1, A1.initialStates);
				A.addEpsTransitions(A1.finalStates, A1.initialStates);
				A.addEpsTransitions(A1.finalStates, Q2);
				A.addEpsTransition(Q1, Q2);
			}
			else {
				XreError.showbug_and_exit("wrong tree passed to the " +
														 "build-automaton algorithm");
			}
		}
		else if (cl == ASTAs.class) {
			if (node.jjtGetNumChildren() == 1) {
				SimpleNode P1 = (SimpleNode)node.jjtGetChild(0);
				int scope = ((ASTAs)node).getScope();
				String name = ((ASTAs)node).getIdentifier();
				SymElement idEl = symtab.search(SymElementKind.VARIABLE,scope,name);
//				SymElement idEl = ((ASTAs)node).getElement();
				
				X.add(idEl);
				TASimple A1 = buildTA(P1,X);
				A.merge(A1, 
					TAPiece.STATES | TAPiece.INITIAL | TAPiece.FINAL | TAPiece.TRANS);
				//X.remove(idEl); // needed?
			}
			else {
				XreError.showbug_and_exit("wrong tree passed to " +
										   "build-automaton algorithm");
			}
		}

		return A;
	}
}