/* 
 * Translate.java
 * project: XRel
 * @author Fabrizio Bisi
 * Created on 24-lug-03
 */
package xrel.analyzer;

import java.util.Vector;

import xrel.XreError;
import xrel.parser.SimpleNode;
import xrel.parser.ASTAs;
import xrel.parser.ASTComma;
import xrel.parser.ASTEmpty;
import xrel.parser.ASTIdentifier;
import xrel.parser.ASTOperator;
import xrel.parser.ASTPar;
import xrel.parser.ASTString;
import xrel.parser.ASTStringLiteral;
import xrel.parser.ASTTag;
import xrel.parser.SpecialNames;
import xrel.parser.XRelParserTreeConstants;
import xrel.parser.SymElementKind;
import xrel.parser.SymElement;
import xrel.parser.SymTable;

/**
 * <p>Provides several transform operations to the symbol table. </p>
 * <p>In details:
 * <ul>
 * <li>addPredefined() adds the predefined type "Any" to the symbol table</li>
 * <li>desugar() performs a few simplifications to all the 
 * expressions in the symbol table (types, patterns, variable types and 
 * expressions)</li>
 * <li>normalization() transforms patterns in an internal format and 
 * prepare them to be transformed in automata</li>
 * </ul></p>
 * <p>To know more see the descriptions of the methods.</p>
 * <p>N.B.: This is a static class so you will never need to allocate an object 
 * of this class.</p>
 * @author Fabrizio Bisi
 */
public class Translate implements XRelParserTreeConstants,SymElementKind,
                                  SpecialNames {
	/**
	 * <p>Adds the predefined type "Any" to the symbol table. 
	 * Remember that Any is defined as in the following:
	 * <blockquote><pre>
	 * typedef Any = (~[Any] | String)*;</pre>
	 * </blockquote></p>
	 * @param st the symbol table
	 */
	public static void addPredefined(SymTable st) {
		/* Note: it's better to consider "Any" as a special node?
		 * Papers talk about this optimization ("A Simpler Design"?)
		 * but it seems that the code don't adopt it (see HType.any).
		 * There is an old version of TAPAT_FULL.PS that talks 
		 * about pattern optimization, refer to this for a widest vision
		 * of the problem.
		 */
		if (st.search(S_TYPES,AnyTypeName) == null) {
			/* Creates the Any node as:
			 * typedef Any = (~[Any] | String)*;
			 * that is:
			 *	Operator *[ Par[ Tag ~[ Identifier "Any" ], String ] ]
			 */
			// create new * operator
			ASTOperator anyNode = new ASTOperator(JJTOPERATOR);
			anyNode.setOperator("*");
			// create new Par (|) node and link it to *
			ASTPar newPar = new ASTPar(JJTPAR);
			anyNode.jjtmAddChildExt(newPar,0);
			// create tag ~ and link it to the Par node
			ASTTag newTag = new ASTTag(JJTTAG);
			newTag.setName(AnyTagName);
			newPar.jjtmAddChildExt(newTag,0); 
			// create identifier Any and link it to the tag ~
			ASTIdentifier newId = new ASTIdentifier(JJTIDENTIFIER);
			newId.setName(0,AnyTypeName);
			newTag.jjtmAddChildExt(newId,0);
			// create the String node and link it to the Par node
			ASTString newStr = new ASTString(JJTSTRING);
			newPar.jjtmAddChildExt(newStr,1);
			// Now add Any to the symbol table
			Vector Types = st.getVector(TYPE);
			SymElement anyType = new SymElement(st,0,AnyTypeName,TYPE,anyNode);
			Types.add(anyType);
		}
	}

	/**
   * <p>Performs a few simplifications to all the elements in the 
   * symbol table: types, patterns, variable types and expressions. </p>
	 * <p>Specifically it performs the following transformations:
	 * <ul>
	 * <li>replaces the ground types (String and string literals) with 
	 * pseudo-tags. The names chosen for pseudo-labels are #String[] for String 
	 * and #".." for string literals. </li>
	 * <li>removes the operators "+" and "?" performing the following 
	 * substitutions:
	 *   <ul>
	 *   <li> P+ = P,P* </li>
	 *   <li> P? = P|()</li>
	 *   </ul>
	 * </li>
	 * </ul>
	 * <p>N.B.:
	 * <ol>
	 * <li>The initial symbol "#" of pseudo-labels grants that it's 
	 * impossibile to have name conflicts with normal labels, as XML tag
	 * names never begin with this symbol (labels in XRel follow the same
	 * syntactic conventions as XML names)</li> 
	 * <li>This function transforms the original syntax tree so that it's
	 * impossible to have a correspondance with tokens</li>
	 * <li>In the transformation of P+ P and its subtree is copied instead of 
	 * referenced twice. This is to have the parent pointer correctly set 
	 * in both the instances.</li>
	 * </ol></p>
	 * @param st the symbol table
	 */
	public static void desugar(SymTable st) {
		// Desugar type expressions
		Vector types = st.getVector(TYPE);
		for (int i=0; i < types.size(); i++) {
			SymElement el = (SymElement)types.elementAt(i);
			el.setNode(desugarElement(st,el.getNode()));
		}
		// Desugar variable types
		/* Commented because they're again linked to the original
		 * syntax tree, so that they're desugared with original patterns 
		 */ 
//		Vector variables = st.getVector(VARIABLE);
//		for (int i=0; i < variables.size(); i++) {
//			SymElement el = (SymElement)variables.elementAt(i);
//			el.setNode(desugarExpression(st,el.getNode()));
//			if (specialChecking) checkDesugaredForm(st);
//		}
		// Desugar patterns
		Vector patterns = st.getVector(PATTERN);
		for (int i=0; i < patterns.size(); i++) {
			SymElement el = (SymElement)patterns.elementAt(i);
			el.setNode(desugarElement(st,el.getNode()));
		}
		/* Also if expressions lack operators we need to desugar them
		 * to remove ground types */
		Vector expressions = st.getVector(EXPRESSION);
		for (int i=0; i < expressions.size(); i++) {
			SymElement el = (SymElement)expressions.elementAt(i);
			el.setNode(desugarElement(st,el.getNode()));
		}
	}
		
	/**
	 * Removes the operators "+" and "?" and replaces ground types (String 
	 * and string literals) with pseudo-labels. </p>
	 * <p>The names chosen for pseudo-labels are #String[] for String and 
	 * #".." for string literals. Note that initial symbol grants that it's 
	 * impossibile to have name conflicts with normal labels, as XML tag
	 * names never begin with "#" symbol (labels in XRel follow the same
	 * syntactic conventions as XML names).</p> 
	 * Operators are removed in the following way:
	 * <ul>
	 * <li> P+ = P,P* </li>
	 * <li> P? = P|()</li>
	 * </ul>
	 * <p>N.B.:
	 * <ol>
	 * <li>This function transforms the original syntax tree so that it's
	 * impossible to have a correspondance with tokens</li>
	 * <li>In the transformation of P+ P and its subtree is copied instead of 
	 * referenced twice. This is to have the parent pointer correctly set 
	 * in both the instances.</li>
	 * </ol></p>
	 * @param st the symbol table
	 * @param node the root node of the subtree to change
	 */ 
	public static SimpleNode desugarElement(SymTable st, SimpleNode node) {
		// go in recursion through the children first
		for (int i = 0; i < node.jjtGetNumChildren(); ++i) {
			SimpleNode n = (SimpleNode)node.jjtGetChild(i);
			if (n != null) {
				SimpleNode newChild = desugarElement(st,n);
				// if the root of the subtree is changed update the children list
				if (n != newChild) {
					node.jjtmAddChildExt(newChild,i);
					// replace the variable pointer in the symbol table
					if (node.getClass() == ASTAs.class) {
						SymElement varEl = ((ASTAs)node).getElement();
						/* I cannot search the variable in the symbol table here 
						 * because in this phase different variables with the same 
						 * name in the same scope could again exist (different 
						 * variables with the same name in the same scope are merged 
						 * in a successive phase): for this I need a pointer to the
						 * correct instance on the ASTAs node */
						if (varEl != null) {
							varEl.setNode(newChild);
							//System.out.println("\n!!!!!!!!!!!!!!!!!!!!!!!!!! changed " + 
							//varEl.getName());
						} 
						else {
							XreError.showerror_and_exit(node + 
								" has not an associated symbol element");
						}
					}
				} 
			} 
		}

		if (node.getClass() == ASTOperator.class) {
			String op = ((ASTOperator)node).getOperator();
			if (node.jjtGetNumChildren() != 1) {
				XreError.showbug_and_exit("[desugar] " + node + 
																	" doesn't have exactly 1 child");
				return null;
			}
			if (op.equals("+")) {
				/* We want to make the following transformation:
				 * +[nodeP] --> comma[nodeP, *[nodeP]]
				 * Warning! node will be duplicated
				 */
				SimpleNode nodeP = (SimpleNode)node.jjtGetChild(0);
				// create comma node 
				ASTComma newcomma = new ASTComma(JJTCOMMA);
				// make the nodeP to be the first child of comma
				node.jjtmMoveChildren(newcomma,0,0);
				// create "operator *" node and link it to the comma as 2nd child
				ASTOperator newstar = new ASTOperator(JJTOPERATOR);
				newstar.setOperator("*");
				newcomma.jjtmAddChildExt(newstar,1);
				/* now create a copy of nodeP (with the whole subtree) and
				 * link it to the "operator *" */
				SimpleNode nodePcopy = nodeP.jjtmCloneSubtree();
				newstar.jjtmAddChildExt(nodePcopy,0);
				// finally returns newcomma as root of the subtree
				node = newcomma;
			} 
			else if (op.equals("?")) {
				/* We want to make the following transformation:
				 * ?[nodeP] --> par[nodeP, empty]
				 */
				ASTPar newpar = new ASTPar(JJTPAR);
				// move subtree from opt to newpar as first child
				node.jjtmMoveChildren(newpar,0,0);
				// add newempty as second child of newpar
				ASTEmpty newempty = new ASTEmpty(JJTEMPTY);
				newpar.jjtmAddChildExt(newempty,1);
				// finally returns newpar as root of the subtree
				node = newpar;
			}
		}
		else if (node.getClass() == ASTString.class) {
			// create a new tag #String[] 
			ASTTag newtag = new ASTTag(XRelParserTreeConstants.JJTTAG);
			newtag.setName(StringTypeName);
			// now create the empty content and link it to the new tag
			ASTEmpty newempty = new ASTEmpty(XRelParserTreeConstants.JJTEMPTY);
			newtag.jjtmAddChildExt(newempty,0);
			// now returns newtag
			node = newtag;
		}
		else if (node.getClass() == ASTStringLiteral.class) {
				String txt = ((ASTStringLiteral)node).getValue();
				// create a new tag #"..."[] 
				ASTTag newtag = new ASTTag(XRelParserTreeConstants.JJTTAG);
				newtag.setName(StrUnquotedPrefix + txt + StrUnquotedSuffix);
				// now create the empty content
				ASTEmpty newempty = new ASTEmpty(XRelParserTreeConstants.JJTEMPTY);
				newtag.jjtmAddChildExt(newempty,0);
				// now returns newtag
				node = newtag;
		}
		
		// return the node itself
		return node;
	}
	
	/**
	 * <p>Internalizes all the patterns of the program, that is it 
	 * transformes them in an internal format so that they can be
	 * easily transformed in automata. </p>
	 * <p>In details it makes the following changes to the patterns:
	 * <ol>
	 * <li>it expands the type names at top-level, i.e. it replaces
	 * type names with theirs related type expressions</li>
	 * <li>it creates new (sub)patterns to group the tag contents</li>
	 * </ol></p>
	 * <p>The resulting patterns have the following syntax:
	 * <blockquote><pre>
	 *   P ::= P as x     variable binder
	 *         l[X]       label
	 *         ()         empty pattern
	 *         P,P        concatenation
	 *         P|P        choice
	 *         P*         repetition</pre>
	 * </blockquote>
	 * where x,.. ranges over variable names and X,.. ranges over pattern
	 * references.
	 * Substantially there are no more type names (because they've been 
	 * expanded) and each tag content corresponds to a specifical pattern. 
	 * These are exactly the changes this function does.</p> 
	 * <p>N.B.:
	 * <ul>
	 * <li>as side effect this function sets the value of the variable
	 * originalPatterns inside the symbol table object. This should be the
	 * unique place in the code where this variable is set</li>
	 * <li>transformations at point 2 can create many new patterns so that 
	 * it should be better to minimize pattern generation (and hence automata
	 * generation) with hashing techniques as XDuce authors do. 
	 * This is left for future work</li>
	 * </ul></p>
	 * @param st the symbol table
	 */
	public static void normalization(SymTable st) {
		// Set st.originalPatterns - I set this variable only here
		st.originalPatterns = st.getVector(PATTERN).size();
		st.originalExpressions = st.getVector(EXPRESSION).size();

		// normalize types
		Vector types = st.getVector(TYPE);
		int k = 0;
		while (k < types.size()) {
			SymElement el = (SymElement)types.elementAt(k);
			if (normalizeElement(st, el, el.getNode(), PATTERN)) k++;
		}
		
		// normalize patterns
		int j = 0;
		Vector patterns = st.getVector(PATTERN);
		while (j < patterns.size()) {
			SymElement el = (SymElement)patterns.elementAt(j);
			/* the loop on each pattern - we advance to the next one
			 * only when the normalization is completed
			 * (if there is a type expansion we need to renormalize the
			 * same pattern) */
			if (normalizeElement(st, el, el.getNode(), PATTERN)) j++;
		}
		
//		// normalize expressions
//		Vector expressions = st.getVector(EXPRESSION);
//		int w = 0;
//		while (w < expressions.size()) {
//			SymElement el = (SymElement)expressions.elementAt(w);
//			if (normalizeElement(st, el, el.getNode(), EXPRESSION)) w++;
//		}
	}
	
	/**
	 * Normalizes the specified symbol table element. 
	 * Call this function in the following way:
	 * 	normalizeElement(st,elem,elem.getNode())
	 * 
	 * Algorithm:
	 * It's recursive. The predicate isNormalized(P) is defined by the following:
	 * <blockquote><pre>
	 *   isNormalized(Empty) = true
	 *   isNormalized(l[X]) =
	 *         l[X]       label
	 *         ()         empty pattern
	 *         P,P        concatenation
	 *         P|P        choice
	 *         P*         repetition</pre>
	 * </blockquote>
	 * 
	 * @param st the symbol table
	 * @param defEl the symbol table element to normalize
	 * @param node the root of the current tree
	 * @return true if the current pattern still needs to be normalized
	 */
	private static boolean normalizeElement(SymTable st, SymElement defEl, 
								SimpleNode node, int subExprKind) {
		Class cl = node.getClass();
	  
	  if (cl == ASTEmpty.class) { 
	  	return true; // nothing to do
	  }
		else if (cl == ASTTag.class) {
			/* This code needs NOT to be before code that handles 
			 * identifier expansion as cl can't have 2 types simultaneously */

		  if (node.jjtGetNumChildren() != 1) {
			  XreError.showbug_and_exit(node + " has a wrong number of children");
			  return true;
			}
			SimpleNode Content = (SimpleNode)node.jjtGetChild(0);
			// special handling of ASTEmpty
			if (Content.getClass() == ASTEmpty.class) {
				SymElement emptyEl = st.search(S_PATTERNS,EmptyPatternName);
				if (emptyEl == null) {
					// create a new pattern
					Content.jjtSetParent(null); // reusing the ASTEmpty node
					emptyEl = new SymElement(st,0,EmptyPatternName,
										PATTERN,Content);
					st.getVector(PATTERN).add(emptyEl);
				}
				// now create a new identifier of type new pattern
				ASTIdentifier newId = new ASTIdentifier(JJTIDENTIFIER);
				newId.setName(emptyEl.getScope(),EmptyPatternName);
				node.jjtmAddChildExt(newId,0);
			}
			// special handling of String (#PCDATA)
			else if (Content.getClass() == ASTTag.class && 
			         ((ASTTag)Content).getName() == StringTypeName) {
				SymElement stringEl = st.search(S_PATTERNS,StringPatternName);
				if (stringEl == null) {
					// create a new pattern
					Content.jjtSetParent(null); // reusing the ASTTag node
					stringEl = new SymElement(st,0, StringPatternName,
										PATTERN,Content);
					st.getVector(PATTERN).add(stringEl);
				}
				// now create a new identifier of type new pattern
				ASTIdentifier newId = new ASTIdentifier(JJTIDENTIFIER);
				newId.setName(stringEl.getScope(),StringPatternName);
				node.jjtmAddChildExt(newId,0);
			}
			else if (Content.getClass() != ASTIdentifier.class) {
				/* TODO when normalizing expressions it can be that there is
				 * l[x] where x is NOT a type but a variable.
				 * What is the correct way to handle this situation?
				 * Currently l[x] doesn't create a new pattern 
				 */
				String ContentStr = Content.getPattern();
				/* this is a little optimization, it doesn't grant that
				 * there will be not duplicates inside the same scoping
				 * (see repetion.xre). The problem is that already
				 * normalized patterns dont' match with original patterns
				 * TODO there is a way to grant uniqueness of structure 
				 * inside same scoping?
				 */
				SymElement pEl = st.searchPattern(ContentStr,defEl.getScope());
				if (pEl == null) {
					// create a new pattern
					pEl =
						new SymElement(st, defEl.getScope(), null, subExprKind, Content);
					// the root of symbol table elements needs to be null
					Content.jjtSetParent(null); 
					st.getVector(subExprKind).add(pEl);
				}
//				else {
//					System.out.println("\n>>>>>>>>>found " + pEl.getUniqueName() + 
//						" in scope " + defEl.getScope() + 
//						" for the string: \"" + ContentStr + "\"");
//				}
				// now create a new identifier node 
				ASTIdentifier newId = new ASTIdentifier(JJTIDENTIFIER);
				newId.setName(pEl.getScope(),pEl.getName());
				node.jjtmAddChildExt(newId,0);
			}
			//else if (Content.getClass() == ASTIdentifier.class) do nothing
			return true;
	  }
		else if (cl == ASTIdentifier.class) { // expand the type identifier
			// this code is never executed when the parent is a tag
			String idName = ((ASTIdentifier)node).getName();
			/* find the identifier. Note that I never need to expand pattern
			 * names as new patterns are created already normalized (i.e. 
			 * they're the unique child inside a tag) so this code is never
			 * executed for them.
			 */
			// search a type name (types always have scope 0)
			SymElement typeEl = st.search(TYPE,0,idName);
			if (typeEl != null) {
				// now replace the identifier with its body
				SimpleNode parent = (SimpleNode)node.jjtGetParent();
				SimpleNode typeNode = typeEl.getNode();
				if (parent == null) {
					// The type name is the root node of the pattern
					/* Copy or reference? Theoretically here I should take the same 
					 * decision as below (where I copy the subtree to mantain coherent 
					 * the parent pointers) but the parent pointer of the root node 
					 * needs to be null so that it should be possible to share a 
					 * subtree for two patterns.
					 * However I copy for uniformity with the decision below
					 */
					SimpleNode typeCopy = typeNode.jjtmCloneSubtree(); // clone type
					defEl.setNode(typeCopy);
				}
				else {
					int idx = parent.jjtmSearchChild(node);
					if (idx >= 0) {
						/* Here I need to copy the whole subtree (if I only copy
						 * a reference the parent pointers from the type subtree to
						 * the patterns will be wrong in all the instances but one) */
						SimpleNode typeCopy = typeNode.jjtmCloneSubtree(); // clone type
						((SimpleNode)parent).jjtmAddChildExt(typeCopy,idx);
						if (parent.getClass() == ASTAs.class) {
							// replace the variable pointer in the symbol table
							SymElement varEl = ((ASTAs)parent).getElement();
							/* I cannot search the variable in the symbol table here 
							 * because in this phase different variables with the same 
							 * name in the same scope could again exist (different 
							 * variables with the same name in the same scope are merged 
							 * in a successive phase): for this I need a pointer to the
							 * correct instance on the ASTAs node */
							if (varEl != null) {
								varEl.setNode(typeCopy);
								//System.out.println("\n!!!!!!!!!!!!!!!!!!!!!!!!!! changed " + 
								//										varEl.getName());
							} 
							else {
								XreError.showerror_and_exit(parent + 
									" has not an associated symbol element");
							}
						}
					}
					else
						XreError.showbug_and_exit("can't find a child in the " + 
																			"children list of its parent");
				}
				// need to normalize again!
				return false;
			} 
			else
				/* TODO we have a variable in an expression, so that we can't 
				 * expand it. How we can handle this situation when we build
				 * the automata?  */
				return true;
		}
	  else if ((cl == ASTComma.class) || (cl == ASTPar.class)) {
		  // "P1,P2" or "P1|P2"
		  boolean rslt = true;
		  for (int i = 0; i < node.jjtGetNumChildren(); i++)
				rslt &=
				 normalizeElement(st, defEl, (SimpleNode)node.jjtGetChild(i),subExprKind);
			return rslt;
	  }
	  else if (cl == ASTAs.class) { 
		  // "P as X" 
		  if (node.jjtGetNumChildren() == 1) 
				return normalizeElement(st, defEl, (SimpleNode)node.jjtGetChild(0),subExprKind);
		  else {
				XreError.showbug_and_exit(node + " doesn't have a child");
			  return true;
		  } 
	  }
	  else if (cl == ASTOperator.class) { 
		  // "P*" ("P+" and "P?" are removed before) 
		  if (node.jjtGetNumChildren() == 1) 
				return normalizeElement(st, defEl, (SimpleNode)node.jjtGetChild(0),subExprKind);
			else {
				XreError.showbug_and_exit(node + " doesn't have a child");
				return true;
			} 
	  }
	  else if ((cl == ASTString.class) || (cl == ASTStringLiteral.class)) {
		  XreError.showbug_and_exit("ground types not fully removed");
		  return true;
	  }
	  else {
			XreError.showbug_and_exit("node type unknown: " + cl);
		  return true;
	  } 
	}
	
	/**
	 * Perform diagnostic checks on the elements inside the symbol table.
	 * <p>The phase parameter can be one of the following:
	 * <dl>
	 * <dt>0</dt>
	 * <dd>called after desugaring</dd>
	 * <dt>1</dt>
	 * <dd>called after normalization</dd>
	 * </dl></p>
	 * @param st the simbol table
	 * @param phase the phase after which is called
	 */
	public static void diagnosticChecks(SymTable st, int phase) {
		Vector types = st.getVector(TYPE);
		for (int i=0; i<types.size();i++) {
			SymElement el = (SymElement)types.elementAt(i);
			switch(phase) {
				case 0: checkDesugaredForm(el.getNode()); break;
				case 1: checkNormalization(st,el.getNode()); break;
				default:
					XreError.showbug_and_exit("unknown diagnostic phase");
			}
		}
		Vector patterns = st.getVector(PATTERN);
		for (int i=0; i<patterns.size();i++) {
			SymElement el = (SymElement)patterns.elementAt(i);
			switch(phase) {
				case 0: checkDesugaredForm(el.getNode()); break;
				case 1: checkNormalization(st,el.getNode()); break;
				default:
					XreError.showbug_and_exit("unknown diagnostic phase");
			}
		}
		Vector variables = st.getVector(VARIABLE);
		for (int i=0; i<variables.size();i++) {
			SymElement el = (SymElement)variables.elementAt(i);
			switch(phase) {
				case 0: checkDesugaredForm(el.getNode()); break;
				case 1: checkNormalization(st,el.getNode()); break;
				default:
					XreError.showbug_and_exit("unknown diagnostic phase");
			}
		}
		Vector expressions = st.getVector(EXPRESSION);
		for (int i=0; i<expressions.size();i++) {
			SymElement el = (SymElement)expressions.elementAt(i);
			switch(phase) {
				case 0: checkDesugaredForm(el.getNode()); break;
				case 1: 
					checkDesugaredForm(el.getNode()); break;
					//checkNormalization(st,el.getNode()); break;
				default:
					XreError.showbug_and_exit("unknown diagnostic phase");
			}
		}
	}

	/** 
	 * This function checks that desugaring and removing of ground types
	 * has been correctly performed.
	 * @param node the root of an expression
	 */
	private static void checkDesugaredForm(SimpleNode node) {
		Class cl = node.getClass();
	  if ((cl == ASTEmpty.class) || (cl == ASTIdentifier.class)) {
	  	if (node.jjtGetNumChildren() > 0)
				XreError.showbug_and_exit("[chkDes] " + node + " has children!");
	  }
		else if (cl == ASTString.class || cl == ASTStringLiteral.class) {
			XreError.showbug_and_exit("[chkDes] Found " + node + " after desugaring");
		}
	  else if ((cl == ASTTag.class) || (cl == ASTAs.class) || 
	  				 (cl == ASTOperator.class) ||
						 (cl == ASTComma.class) || (cl == ASTPar.class)) {
			// signal wrong number of children
			if (((cl == ASTTag.class) || (cl == ASTAs.class) || 
			    (cl == ASTOperator.class)) && node.jjtGetNumChildren() != 1)
				XreError.showbug_and_exit("[chkDes] " + node + " doesn't have 1 child");
			// signal presence of wrong operators
			if (cl == ASTOperator.class) {
				String op = ((ASTOperator)node).getOperator();
				if (!op.equals("*"))
					XreError.showbug_and_exit(node + " found after desugaring");
			}
			// recursion on children
			for (int i=0; i< node.jjtGetNumChildren(); i++)
				  checkDesugaredForm((SimpleNode)node.jjtGetChild(i));
	  }
	  else
	  	XreError.showbug_and_exit("[chkDes] unknown node kind:" + cl);
	}
	
	// check normalization and the structure
	static private void checkNormalization(SymTable st, SimpleNode node) {
		Class cl = node.getClass();
	  
		if (cl == ASTTag.class) {
		  if (node.jjtGetNumChildren() != 1)
			  XreError.showbug_and_exit("[chkNor] " + node + 
																	" has a wrong number of children");
			SimpleNode Content = (SimpleNode)node.jjtGetChild(0);
			if (Content.getClass() != ASTIdentifier.class)
				XreError.showbug_and_exit("[chkNor] " + node + 
									" doesn't have an identifier as child after normalization");
	  }
		else if (cl == ASTIdentifier.class) { 
			if (node.jjtGetNumChildren() > 0)
				XreError.showbug_and_exit("[chkNor] " + node +  " has children!");
			SimpleNode parent = (SimpleNode)node.jjtGetParent();
			if (parent.getClass() != ASTTag.class) {
				String idName = ((ASTIdentifier)node).getName();
				// search a type name (types always have scope 0)
				SymElement typeEl = st.search(TYPE,0,idName);
				if (typeEl != null)
					XreError.showbug_and_exit("[chkNor] " + node + 
																" has not been expanded by the normalization");
			}
		}
		else if (cl == ASTEmpty.class) {
			if (node.jjtGetNumChildren() > 0) 
				XreError.showbug_and_exit("[chkNor] empty node has children!");
		}
		else if ((cl == ASTString.class) || (cl == ASTStringLiteral.class)) {
			XreError.showbug_and_exit("[chkNor] ground types not fully removed");
		}
	  else if ((cl == ASTOperator.class) || (cl == ASTAs.class) || 
					  (cl == ASTComma.class) || (cl == ASTPar.class)) {
			if (((cl == ASTAs.class) || (cl == ASTOperator.class)) && 
					node.jjtGetNumChildren() != 1)
				XreError.showbug_and_exit("[chkNor] " + node + 
																	" has a wrong number of children)");
			// recursion on children
			for (int i = 0; i < node.jjtGetNumChildren(); i++)
				checkNormalization(st,(SimpleNode)node.jjtGetChild(i));
	  }
	  else
			XreError.showbug_and_exit("[chkNor] unknown node kind:" + cl);
	}
}
