/*
 * WellFormedness.java
 * project: XRel
 * @author Fabrizio Bisi
 * Created on 15-lug-03
 */
package xrel.analyzer;

import java.util.HashSet;

import xrel.XreError;
import xrel.parser.ASTImportSt;
import xrel.parser.SpecialNames;
import xrel.parser.SymElementKind;
import xrel.parser.SymElement;
import xrel.parser.SymTable;
import xrel.parser.SimpleNode;
import xrel.parser.Node;
import xrel.parser.ASTAs;
import xrel.parser.ASTComma;
import xrel.parser.ASTEmpty;
import xrel.parser.ASTIdentifier;
import xrel.parser.ASTOperator;
import xrel.parser.ASTPar;
import xrel.parser.ASTString;
import xrel.parser.ASTStringLiteral;
import xrel.parser.ASTSwitchCase;
import xrel.parser.ASTTag;
import xrel.parser.ASTTypeDecl;

/**
 * Provides some well-formedness checks on type expressions and
 * patterns. Specifically:
 * <ul>
 * <li> disconnectedness allows to check there isn't recursion at the 
 *      top-level inside the type declarations of the program </li>
 * <li> checkLinearity checks that all the patterns inside the program
 *      are linear. </li>
 * </ul>
 * <p>N.B.: This is a static class, you will never need to allocate an object 
 * of this class.</p>
 * @author Fabrizio Bisi
 */
public class WellFormedness implements SpecialNames {
	private static SymTable symtab = null;

	/**
	 * <p>Checks that all the type declarations in the program don't have 
	 * recursion at the top-level. For example the following definitions 
	 * are wrong:
	 * <blockquote><pre>
	 * typedef X = a[],X | ();     // self-recursive
	 * typedef Y = b[],Z | String; // mutually recursive with Z
	 * typedef Z = c[],Y | String; //    ''        ''    with Y</pre>
	 * </blockquote>
	 * while the following recursive definitions are correct:
	 * <blockquote><pre>
	 * typedef A = a[B] | String;
	 * typedef B = b[A] | String;</pre>
	 * </blockquote>
	 * <p>N.B.:<ul><li> 
	 * This function doesn't work when you call it on other nodes than root
	 * </li></ul></p>
	 * <p><b>Algorithm:</b></p>
	 * <p>Let x,y,.. range over type names and let T,U,.. range over type 
	 * expressions.
	 * For each type declaration in the program:</p>
	 * <blockquote><pre>
	 * typedef x = T;</pre>
	 * </blockquote>
	 * <p>we need to check that x don't recur neither directly inside T
	 * neither inside the expansion of a type name reachable at top-level
	 * from T.
	 * For expansion of a type name y we intend its related type expression
	 * U in the statement "typedef y = U".
	 * A type name y is reachable at top-level from a type expression T if 
	 * it's possible to reach y from T by subsequent expansions. 
	 * For example in: 
	 * <blockquote><pre>
	 * typedef x = y;
	 * typedef y = z;</pre>
	 * </blockquote>
	 * both y and z are reachable from the expansion of x.</p>
	 * <p>For each type declaration "typedef x = T" the main function calls 
	 * a subroutine "dc" with parameters dc(emptySet,T,x) that checks that x 
	 * is not reachable at top-level from T.</p>
	 * <p>Here's a pseudo-code for the subroutine:</p>
	 * <blockquote><pre>
	 * boolean dc(sigma,T,x) {
	 *   switch (kind of T) {
	 *     case "|", ",": 
	 *       return dc(sigma,T.child(1),x) && .. && dc(sigma,T.child(n),x);
	 *     case empty,tag,literal: 
	 *       return true;         // base case 1 (success)
	 *     case x: return false;  // base case 2 (fails!)
	 *     case y != x: 
	 *       if (y is in sigma) return true;
	 *       else return dc(sigma + {y},U,x); // "typedef y = U;"
	 *   }
	 * }</pre>
	 * </blockquote>
	 * <p>The first argument "sigma" of the function is the set of type 
	 * names we have already expanded (initially empty). We return 
	 * immediately true when we found a string literal, an empty node or a 
	 * tag (remember that we're searching only top-level recursion); 
	 * conversely we return false when we found x. 
	 * If we encounter another type name y different from x we return true 
	 * if it's present in sigma (as this means we have already checked it), 
	 * otherwise we recursively call dc with sigma = sigma + {y} checking 
	 * if x is in the expansion of y.</p>
	 * <p>Note that we don't need to check the export type as it
	 * can't be recursive.</p>
	 * @param st the symbol table
	 * @param root the root node of the syntax tree
	 * @return true if all the type declarations are disconnected, 
	 * false otherwise
	 */
	public static boolean disconnectedness(SymTable st, SimpleNode root) {
		symtab = st;
		boolean rslt = true;
		
		/* check all the typedef statements */
		int numChildren = root.jjtGetNumChildren();
		for (int i = 0; i < root.jjtGetNumChildren(); i++) {
			SimpleNode tydec = (SimpleNode)root.jjtGetChild(i);
			// check only typedef statements (not the export clause!)
			if (tydec.getClass() == ASTTypeDecl.class) {
				if (tydec.jjtGetNumChildren() == 1) {
					String x = ((ASTTypeDecl)tydec).getIdentifier();
					SimpleNode expr = (SimpleNode)tydec.jjtGetChild(0);
					HashSet sigma = new HashSet();
					// check if the type declaration is disconnected
					rslt &= dc(sigma,expr,x);
				}
				else {
					XreError.showbug_and_exit("malformed type declaration");
					return false;
				}
			}
			/* Warning! this code works because typedefs are the first statements
			 * of the program!
			 */
			else break;
		}

		return rslt;
	}

	/**
	 * <p>Checks that all the patterns inside the program are linear. Remember
	 * that a pattern is linear if variables inside it are bound exactly once.
	 * </p>
	 * <p>N.B.:
	 * <ul><li> 
	 * This function doesn't work when you call it on other nodes than root
	 * </li></ul></p>
	 * <p><b>Algorithm:</b></p>
	 * <p>Let BV(P) the set of variables bound in the pattern P. For example
	 * in the following pattern P:
	 * <blockquote><pre> 
	 * l1[T as x], (l2[] as y | l3[U as y])</pre>
	 * </blockquote>
	 * BV(P) = {x,y}</p>
	 * <p>Note that BV(P) is similar to the set of reachable type names at 
	 * top-level in a type expression T (let we call it reach(T) now) with 
	 * the following main differences:
	 * <ul>
	 * <li>BV(P) is a set of variable names while reach(T) is a set of 
	 * type names</li>
	 * <li>BV(P) collect also bindings inside tags, reach(T) ignores 
	 * type names inside tags</li>
	 * <li>reach(T) collects type names reachable from type name expansions,
	 * BV(P) don't need to expand type names as type expressions don't
	 * have bindings</li>
	 * </ul></p>
	 * For each pattern in the program (that of the import statement
	 * and patterns of the clauses in the typeswitch construct) we 
	 * recursively check the followings:
	 * <blockquote><pre>
	 * P = P1 as x     x is not in P1
	 * P = P1,P2       intersection(BV(P1),BV(P2)) = the empty set
	 * P = P1|P2       BV(P1) = BV(P2)
	 * P = P1*         BV(P1) = the empty set</pre>
	 * </blockquote></p>
	 * @param st the symbol table
	 * @param root the root node of the syntax tree
	 * @return true if all the patterns are linear, false otherwise
	 */
	public static boolean checkLinearity(SymTable st, SimpleNode root) {
		symtab = st;
		boolean rslt = true;
		
		/* Skip all type declarations and finds the import and the
		 * typeswitch statements.
		 * Warning! this code functions only with the following structure
		 * of the program:
		 * Program ::= 	TypeDef* ImportSt ExportTp? TypeSwitch?
		 * because it checks only the last 3 top-level statements.
		 */ 
		int numChildren = root.jjtGetNumChildren();
		SimpleNode typeswitch= null, importst= null;
		for (int i = Math.max(0,numChildren-3); i<numChildren; i++) {
			SimpleNode curr = (SimpleNode)root.jjtGetChild(i);
			if (curr.getClass() == ASTImportSt.class) importst = curr;
			else if (curr.getClass() == ASTSwitchCase.class) typeswitch = curr;
		}
		// error checking (this is a bug!)
		if (importst == null)
			XreError.showbug_and_exit("import statement missing");
		
		/* Now call check_pat_linearity on the mandatory import statement
		 * and on each clause of the optional typeswitch statement
		 */ 
		rslt = check_pat_linearity(importst.jjtGetChild(0));
		if (typeswitch != null)
			for (int i = 0; i < typeswitch.jjtGetNumChildren(); i++) {
				// check however all the clauses to collect errors
				rslt &= 
					check_pat_linearity(typeswitch.jjtGetChild(i).jjtGetChild(0));
			}
		return rslt;
	}
	
	/* Checks if a type expression is disconnected from the type name x,
	 *  i.e if x does not occur in the top-level of the expression
	 * (check recursively inside other type definitions).
	 * For algorithm details see the class entry point
	 */
	private static boolean dc(HashSet sigma,SimpleNode node, String x) {
		Class cl = node.getClass();
		  
	  if ( (cl == ASTEmpty.class)  || (cl == ASTTag.class) ||
		     (cl == ASTString.class) || (cl == ASTStringLiteral.class) )
			/* base case 1 (success): (),String,"...",l[...]
			 * N.B. x represents a type name, not a pattern, so that 
			 * it can't have internal bindings.
			 */
			return true; 
		else if (cl == ASTIdentifier.class) {
			String y = ((ASTIdentifier)node).getName();
			// if (y==x) fails!
			if (y.equals(x)) {
				symtab.add_type_error((SimpleNode)node,
					"top-level recursion of identifier " + x);
				return false;
			} 
			// if (y != x && y is in sigma) return true;
			else if (sigma.contains(y)) return true;
			else {
				/* if (y != x && y is not in sigma)
				 *    return dc(sigma + {Y},expand(Y),X); // typedef Y = expand(Y)
				 */
				sigma.add(y);
				SimpleNode y_node;
				SymElement y_elem = symtab.search(SymElementKind.TYPE,0,y);
				if (y_elem != null) y_node = y_elem.getNode();
				else {
					XreError.showbug_and_exit("identifier " + y + "not found");
					return false; 
				}
				boolean rslt = dc(sigma,y_node,x);
				sigma.remove(y); // really needed?
				return rslt;
			}
		}
		else if ((cl == ASTComma.class) || (cl == ASTPar.class)) {
			//		case "|", ",": 
			//		  return dc(sigma,child 1,X) && .. && dc(sigma,child n,X);
		  boolean rslt = true;
		  
			for (int i= 0; i < node.jjtGetNumChildren(); i++)
			  rslt &= dc(sigma,(SimpleNode)node.jjtGetChild(i),x);
		  return rslt;
		}
		else if (cl == ASTAs.class) { 
			XreError.showerror_and_exit("I can't check disconnectedness of patterns");
		  return false;
	  }
	  else if (cl == ASTOperator.class) { 
		  boolean rslt1;
		  if (node.jjtGetNumChildren() == 1) 
			  return dc(sigma,(SimpleNode)node.jjtGetChild(0),x);
		  else {
				XreError.showbug_and_exit("malformed operator inside tree");
				return false;
			}
	  }
	  else {
			XreError.showbug_and_exit("node type unknown");
		  return false;
	  } 
	}

	/*
	 * check if a pattern is linear, i.e. if all variables bind exactly once.
	 * For algorithm details see the class entry point
	 */
	private static boolean check_pat_linearity(Node node) {
		HashSet emptySet = new HashSet();
		
		return check_pat_rec_linearity(emptySet,node);
	}

	/* Checks if a pattern is linear, i.e. if all variables bind 
	 * exactly once. */
	private static boolean check_pat_rec_linearity(HashSet bvar, Node node)  {
	  Class cl = node.getClass();
	  
		if ( (cl == ASTEmpty.class) || 
		     (cl == ASTString.class) || (cl == ASTStringLiteral.class) || 
				 (cl == ASTIdentifier.class) ) 
			/* base cases ((),String,"...",X are linear)
			 * N.B. X represents a type name, not a pattern name, so
			 * that it can't have internal bindings. In the final algorithm
			 * we will need to go recursively inside patterns 
			 */
	  	return true; 
	  else if (cl == ASTTag.class) {
			if (node.jjtGetNumChildren() == 1) 
	  		return check_pat_rec_linearity(bvar, node.jjtGetChild(0));
	  	else return true; // if == 0
	  }
		else if (cl == ASTComma.class) {
			// given "P1,P2" check that P1 and P2 have disjoint variables
			boolean rslt = true;
			boolean isLinear = true;
		  
		  /* return true if all the children are linear and their 
		   * bindings are disjoint.
		   * Also when it finds an error this function continues to scan
		   * the rest of the subtrees to collect other errors
			 */
		  for (int i= 0; i < node.jjtGetNumChildren(); i++) {
				HashSet bvar1 = new HashSet();
				rslt &= check_pat_rec_linearity(bvar1,node.jjtGetChild(i));
				if (isLinear)
					/* check if tyenv and tyenv1 are disjoint and if is this the 
					 * case fill join them */
					isLinear &= unionDisjoint(bvar,bvar1);
		  }
		  
		  if (isLinear) return rslt;
		  else {
				symtab.add_type_error((SimpleNode)node,
					"linearity constraint broken (subsequent redefinitions)");
		  	return false;
		  }
		}
		else if (cl == ASTPar.class) {
			// given "P1|P2" check that P1 and P2 have the same variables 
			boolean rslt = true;
			boolean isLinear = true;
			/* return true when all children are linear and have the same
			 * bindings
			 */
			if (node.jjtGetNumChildren() > 0) {
				/* fill tyenv with the first tyenv1 */
				HashSet bvar1 = new HashSet();
				rslt = check_pat_rec_linearity(bvar1,node.jjtGetChild(0));
				bvar.addAll(bvar1);
			}
			else
				XreError.showbug_and_exit("malformed tree");

			for (int i= 1; i < node.jjtGetNumChildren(); i++) {
				HashSet bvar1 = new HashSet();
			  rslt &= check_pat_rec_linearity(bvar1,node.jjtGetChild(i));
			  if (isLinear)
					/* check if tyenv and tyenv1 are equal and if it's not the 
					 * case signal error */
				  isLinear &= bvar.equals(bvar1);
			}
			if (isLinear) return rslt;
			else {
				symtab.add_type_error((SimpleNode)node,
					"linearity constraint broken (variables in all choices need to be equal)");
				return false;
			}
		}
		else if (cl == ASTAs.class) { 
			/* Given "P as x" check that x is not bound again inside P.
			 * Furthermore it returns back tyenv = union(env(P),{X})
			 */
			HashSet bvar1 = new HashSet();
			boolean rslt;

			if (node.jjtGetNumChildren() == 1) 
				rslt = check_pat_rec_linearity(bvar1,node.jjtGetChild(0));
			else {
				XreError.showbug_and_exit("malformed binding in a pattern");
				return false;
			} 
			String newid = ((ASTAs)node).getIdentifier();
			boolean isLinear = !bvar1.contains(newid);
			if (isLinear) {
				if (rslt) {
					bvar.addAll(bvar1);
					// add bound variable to tyenv
					bvar.add(newid);
				}
				return rslt;
			}
			else {
				symtab.add_type_error((SimpleNode)node,
					"linearity constraint broken (nested variable definition)");
				return false;
			}
		}
		else if (cl == ASTOperator.class) { 
			/* Given "P*" or "P+" or "P?" check that P is linear and that
			 * inside P there are not bindings
			 */
			HashSet bvar1 = new HashSet();
			boolean rslt;
			if (node.jjtGetNumChildren() == 1) 
				rslt = check_pat_rec_linearity(bvar1,node.jjtGetChild(0));
			else {
				XreError.showbug_and_exit("malformed tree");
				rslt = false; // for compiler
			}
			boolean isLinear = bvar1.isEmpty(); 
			if (isLinear) {
				if (rslt) bvar.addAll(bvar1);
				return rslt;
			}
			else {
				symtab.add_type_error((SimpleNode)node,
					"linearity constraint broken (can't bind inside operators *,+,?)");
				return false;
			} 
		}
		else {
			XreError.showbug_and_exit("node type unknown");
			return false;
		} 
	}
	
	/* Union between disjoint sets
	 * add tyenv2 to tyenv1 and return true if there are not common elements
	 * (if common elements are found tyenv1 is partially filled with elements
	 *  by tyenv2 and the function returns false)
	 */
	private static boolean unionDisjoint(HashSet tyenv1, HashSet tyenv2) {
		boolean disj = true;
		java.util.Iterator it = tyenv2.iterator();
		while (it.hasNext()) {
			String id = (String)it.next();
			if(! tyenv1.add(id)) {
				disj = false;
				break;
			}
		}
		return disj;
	}
}
