/*
 * CheckVisitor.java
 * project: XRel
 * @author Fabrizio Bisi
 * Created on 13-giu-03
 */
package xrel.parser;

import xrel.XreError;

/**
 * This is a JJTree visitor that checks for syntax errors in the 
 * syntax tree of the program.
 * <p>Here's a list of the collected errors:
 * <ul>
 * <li>detection of unknown identifiers</li>
 * <li>check that variables are not used in type declarations 
 * (typedef/export)</li>
 * <li>check that variables are only defined and not used in patterns</li>
 * <li>check that the identifier used in the typeswitch statement is a
 * variable declared in the import clause</li>
 * <li>check that identifiers in code blocks are variables declared
 * in the import clause or in the current clause</li>
 * <li>check that there are not bindings in code blocks</li>
 * </ul></p> 
 * <p>A list of errors that are NOT collected from this visitor:
 * <ul>
 * <li>check that variables are not defined in type declarations
 * (checked at parsing phase)</li>
 * <li>check that an identifier bound in a pattern is not a type
 * name (checked at parsing phase)</li>
 * <li>redefinition of Any type (checked at parsing phase)</li>
 * <li>Warning: global variable hidden</li>
 * <li>all the linearity constraints for patterns</li>
 * <li>check that there is not recursion at top-level for types</li>
 * <li>all the semantics checks (irredundancy and exhaustivity of the 
 * clauses, adherence to the output type etc.)</li>
 * </ul></p>
 * <p>N.B.:
 * <ul>
 * <li>Why to use JJTree visitors instead of normal tree visits?
 * When new node types are added (this happens  almost every time 
 * the grammar is changed) JJTree visitors give errors, so that 
 * it's easy to understand where code needs changes. 
 * JJTree documentation says this is a feature: in my opinion this 
 * is the <u>unique</u> feature, however this feature is essential. 
 * The drawback is that visitors are less easy to read and to 
 * understand.</li>
 * </ul></p>
 * @see xrel.analyzer.WellFormedness
 * @see xrel.analyzer.SemanticChecks
 * @author Fabrizio Bisi
 */

public class ErrorCheckVisitor  implements XRelParserVisitor, SymElementKind {
	final static int 
		KND_TYPEDECL = 1,
		KND_PATTERN = 2,
		KND_BLOCK = 3,
		KND_SWITCHID = 4;
	
	protected int ExprKind = 0;
	protected SymTable symtab;
	
	/**
	 * Builds the object.
	 * @param s the symbol table
	 */
	public ErrorCheckVisitor(SymTable s)
	{
		  symtab = s;
	}
	
	/** Default visitor. It should never be called. */
	public Object visit(SimpleNode node, Object data) {
		XreError.showbug_and_exit(node + 
							": acceptor not unimplemented in subclass?");
/*
	  data = node.childrenAccept(this, data);
 */
	  return data;
	}

	public Object visit(ASTTypeDecl node, Object data) { 
		ExprKind = KND_TYPEDECL; 
		checkId(node,node.getIdentifier(),0,node.getIdToken());
		return check(node,data); 
	}
	public Object visit(ASTImportSt node, Object data) { 
		ExprKind = KND_PATTERN; 
		return check(node,data); 
	}
	public Object visit(ASTExportTp node, Object data) { 
		ExprKind = KND_TYPEDECL; 
		return check(node,data); 
	}
	public Object visit(ASTSwitchCase node, Object data) { 
		// needed for the identifier in the switch statement 
		ExprKind = KND_SWITCHID; 
		checkId(node,node.getIdentifier(),0,node.getIdToken());
		return check(node,data); 
	}
	public Object visit(ASTCase node, Object data) { 
		if (node.children != null && node.children.length == 2) {
			ExprKind = KND_PATTERN;
			node.children[0].jjtAccept(this, data);
			ExprKind = KND_BLOCK; 
			node.children[1].jjtAccept(this, data);
		}
		else
			XreError.showbug_and_exit("malformed tree");
		return data;
	}
	public Object visit(ASTStatementList node, Object data) { 
		ExprKind = KND_BLOCK;
		return check(node,data); 
	}
	public Object visit(ASTAs node, Object data) { 
		checkId(node,node.getIdentifier(),node.getScope(),node.getIdToken());
		return check(node,data); 
	}
	public Object visit(ASTIdentifier node, Object data) {
		checkId(node,node.getName(),node.getScope(),node.first);
		return check(node,data);
	}
	public Object visit(ASTProgram node, Object data) { return check(node,data); }
	public Object visit(ASTPar node, Object data) { return check(node,data); }
	public Object visit(ASTComma node, Object data) { return check(node,data); }
	public Object visit(ASTOperator node, Object data) { return check(node,data); }
	public Object visit(ASTTag node, Object data) { return check(node,data); }
	public Object visit(ASTEmpty node, Object data) { return check(node,data); }
	public Object visit(ASTString node, Object data) { return check(node,data); }
	public Object visit(ASTStringLiteral node, Object data) { return check(node,data); }
	
	/**
	 * Checks for correct use of an identifier.
	 * @param node the current node
	 * @param idName the identifier name
	 * @param idScope the identifier scope (_0 for typedef/import/export,
	 * _n for the n-th clause of typeswitch)
	 * @param idToken the token related to the identifier
	 */
	private void checkId(SimpleNode node, String idName, int idScope, 
											Token idToken) {
		int idKind;
			
		/* Searches for a variable or a type with name idName.
		 * I don't search for patterns and expressions as they have a
		 * name assigned by the program.
		 */
		SymElement el = symtab.search(S_VARIABLES | S_TYPES, idName);
		// idKind = VARIABLE, TYPE or UNDEFINED
		if (el != null) idKind = el.getKind();
		else idKind = UNDEFINED;
			
		/* ExprKind gives the context and it's set before to call this
		 * function.
		 * Here's the values: 
		 * - TYPEDECL: we are in a type declaration (typedef/export)
		 * - PATTERN: we are inside a pattern (case/import)
		 * - SWITCHID: we are in the typeswitch node
		 * - BLOCK: we are inside a block of code (case)
		 */
		switch (ExprKind) {
			case KND_TYPEDECL: // we are in a type declaration
				/* Note: I don't signal error when there is a binding
				 * because this error is already signaled in the parsing phase
				 */
				if ((idKind != TYPE) && node.getClass() != ASTAs.class)
					// signal both for unknown ids and for wrong use of a variable
					signalError(idName,idToken,"% is not a type name");
				break;
			case KND_PATTERN: // we are in a pattern
				if (idKind == TYPE) {
					if (node.getClass() == ASTAs.class) {
						/* we won't signal this error as is already been signaled
						 * in the parsing phase */
						//signalError(id,idToken,,"% is a type, " + 
						//						"you can't bind a pattern with a type name");
					}
				} 
				else if (idKind == VARIABLE) {
					if (node.getClass() != ASTAs.class) 
						signalError(idName,idToken,"% is a variable, " + 
												"you can't use variables in patterns");
				}
				else signalError(idName,idToken,"% is unknown");
				break;
			case (KND_SWITCHID): // the control for the switch id.
				if (node.getClass() == ASTSwitchCase.class) {
					if (idKind == TYPE) { 
						signalError(idName,idToken,"% is a type name, can't match");
					}
					else if (idKind == VARIABLE) {
						// perform an exact search (a variable in the import clause)
						SymElement exact =
						 symtab.search(VARIABLE,0,idName);
						if (exact == null) {
							signalError(idName,idToken,
								"variable % undeclared in this context");
						}
					}
					else signalError(idName,idToken,"% is unknown");
				}
				else {
					XreError.showbug_and_exit("#SWITCHID");
					//throw new Exception("XRel internal bug #1");
				}
				break;
			case (KND_BLOCK): // we are in a block
				if (idKind == TYPE) {
					signalError(idName,idToken,
            "% is a type name, you can't use types in a block");
				} 
				else if (idKind == VARIABLE) {
					if (node.getClass() == ASTAs.class)
					 signalError(idName,idToken,"attempt to use % for binding, " +
										   "you can't bind in a block");
					else {
						// perform an exact search (the same scope or the import clause)
						SymElement exact =
						 symtab.search(VARIABLE,idScope,idName);
						if (exact == null) 
							exact = symtab.search(VARIABLE,0,idName);
						if (exact == null) {
							signalError(idName,idToken,
								"variable % undeclared in this context");
						}
					 }
				}
				else signalError(idName,idToken,"% is unknown");
				break;
			default:
				XreError.showbug_and_exit("#STREXPR");
				//throw new Exception("XRel internal bug #2");
				break;
		}
	}
	
	/**
	 * Recursevely propagates to the children
	 * @param node the current node
	 * @param data unused
	 * @return the same received object
	 */
	private Object check(SimpleNode node, Object data) {
		node.childrenAccept(this,data);
	  return data;
	}

	/**
	 * Signals a syntax error. If in the error string there is "%"
	 * it will be replaced by the name of the identifier node
	 * @param idName the name of the identifier
	 * @param idToken the token associated to the identifier
	 * @param msg the error message
	 */
	private void signalError(String idName, Token idToken, String msg)
	{
		// Replaces the first instance of % with idName.
		// Note that in Java 1.4 it should be: pos = msg.replaceAll("%",idName);
		int pos = msg.indexOf("%");
		if (pos > -1) {
			String prefix = pos-1 >= 0 ? msg.substring(0,pos) : "";
			String postfix = pos+1 < msg.length() ? 
												msg.substring(pos+1,msg.length()) : "";
			msg = prefix + idName + postfix; 
		}
		// now insert the error inside the symbol table
		symtab.add_syntax_error(idToken, idToken, msg);	
	}
}
