/*
 * SymElement.java
 * project: XRel
 * @author Fabrizio Bisi
 * Created on 12-giu-03
 */
package xrel.parser;

import xrel.XreError;
import xrel.analyzer.TA;
import xrel.analyzer.TASimple;

/**
 * This class is a symbol table element. A symbol table element can be 
 * a type name, a pattern, a variable or an expression (expressions
 * are the parameters of printf statements).
 * @author Fabrizio Bisi
 */
public class SymElement {
	private SymTable symtab = null; // I'm using this pointer for automata
	private static int
		typeCounter = 0,
		patternCounter = 0,
		expressionCounter = 0,  
		varCounter = 0; // used for incremental variable id
	private int idCode; // code that identifies univocally the element 
	private String name;
	private int kind; // type, variable, pattern 
	private SimpleNode node; // the poiinter to the original syntax tree
	private SimpleNode value; // the value for the variables
	private int env; // scoping
	// quick reference to the respective automaton (only vars & patterns)
	private TASimple automaton;
	private TA inferredAutomaton;
	
	/**
	 * This is the only constructor you can use to initialize an object
	 * of the symbol table. Usually the symbol elements are built from
	 * the parser when it builds the symbol table, however there are
	 * a few functions that manipulate the expression trees and that 
	 * use this constructor.
	 * Name is always computed for expressions so that the given name is 
	 * ignored, therefore it's possible to pass a null pointer. 
	 * For patterns if the caller gives a null name the name is
	 * computed, otherwise is used the given name. The name of types and 
	 * variables is always specified by the caller and it can't be null. 
	 * @param st the symbol table
	 * @param scope the integer representing the scope of the element
	 * @param nm the name of the element
	 * @param k the kind of the element (type, variable, pattern, expression)
	 * @param nd the root of the subtree associated to the element
	 */
	public SymElement(SymTable st, int scope, String nm, int k, SimpleNode nd) {
		symtab = st;
		env = scope;
		kind = k;
		if ((k < 2) && (k > 5))
			XreError.showbug_and_exit("#STKND");
		node = nd;
		value = null;      // run-time value for variables
		automaton = null;  // quick reference to the automaton
		switch (k) {
			case SymElementKind.VARIABLE: 
				idCode = varCounter++; 
				name = nm; 
				break;
			case SymElementKind.TYPE: 
				idCode = typeCounter++; 
				name = nm; 
				break;
			case SymElementKind.PATTERN: 
				idCode = patternCounter++;
				if (nm == null) name = SpecialNames.PatternPrefix + idCode;
				else name = nm;
				break;
			case SymElementKind.EXPRESSION: 
				idCode = expressionCounter++; 
				name = SpecialNames.ExpressionPrefix + idCode;
				break;
			default: 
				XreError.showbug_and_exit("unknown element kind");
		}
	}
	
	/**
	 * Gets the name of the element.
	 * @return the name
	 */
	public String getName() {
		return name;
	}
	/**
	 * <p>Returns the code for the element. </p>
	 * <p>It's an incremental number for patterns and variables. </p> 
	 * @return the unique identifier code for the element
	 */
	public int getCode() {
		return idCode;
	}
	/**
	 * For variables returns the name with a unique identifier.
	 * For other elements is the same as getName();
	 * @return the unique name of the element
	 */
	public String getUniqueName() {
			return "#" + idCode + ": " + name;
	}
	/**
	 * Gets the subtree of the syntax tree associated to the element.
	 * @return the root node of the subtree
	 */
	public SimpleNode getNode() {
		return node;
	}
	/**
	 * Sets the subtree associated to the element.
	 * @param n the new root node for the variable
	 */
	public void setNode(SimpleNode n) {
		if (node != n) {
			node = n;
			// n.parent = null; // null parent!
		}
	}
	/**
	 * Gets the value associated to this variable after a binding.
	 * @return the root node of the value or null
	 */
	public SimpleNode getValue() {
		return value;
	}
	/**
	 * Sets the value associated to this variable as a result of a binding.
	 * @return the root node of the value or null
	 */
	public void setValue(SimpleNode v) {
		value = v;
	}
	/**
	 * Gets the kind of the element. It can be VARIABLE, TYPE,
	 * PATTERN or EXPRESSION.
	 * @return an integer that represents the kind of the element
	 */
	public int getKind() {
		return kind;
	}
	/* keep care that setKind should change automaton */
	/**
	 * Gets the scope of the element. The scope of an element is the piece
	 * of the code where you can use the element and in this simple language
	 * is represented by an integer: always 0 (global) for types, for 
	 * variables is the number of the "case" (from 1) where they're defined 
	 * (0 if they're defined in the import statement). The scope value
	 * of a pattern and of an expression identifies the pattern or the expression
	 * itself and it has the same numeration conventions as variables 
	 * (therefore the expression with scope 0 doesn't exist as the first clause 
	 * has scope 1).  
	 * @return an integer that represents the scope of a variable or 0 for a 
	 * type or an identifier for a pattern or an expression
	 */
	public int getScope() {
		return env;
	}
	/**
	 * Gets the name of the scope, that is "document" for the import statement
	 * and "case n" for the n-th clause in the typeswitch instruction.
	 * @return the name of the scope
	 */
	public String getScopeStr() {
		if (env == 0) return "document";
		else return "case " + env;
	}
	
	public TA getInferredType() {
		return inferredAutomaton;
	}
	
	public void setInferredType(TA aut) {
		inferredAutomaton = aut;
	}
	
	public String getAutomatonName() {
		String name;
		switch (kind) {
			case SymElementKind.PATTERN: name = "p"; break; 
			case SymElementKind.VARIABLE: name = "v"; break;
			// Warning! only output type can have an automaton!
			case SymElementKind.TYPE: name = "t"; break;
			case SymElementKind.EXPRESSION: name = "e"; break;
			default: 
				XreError.showbug_and_exit("unknown element kind");
				return null;
		}
		name += this.getCode();
		return name;
	}
	
	
	public boolean hasAutomaton() {
		return automaton != null;
	}
	/**
	 * Gets the internal automaton associated to the element and if it
	 * doesn't exist it creates it. 
	 * If this method is invoked on a symbol table element that can't have 
	 * an automaton it returns null.
	 * @return the automaton of a symbol table element or null
	 */
	public TASimple getElementAutomaton(int BuildPhases, boolean debug) {
		if (automaton == null) {
			automaton = 
			  symtab.getSimpleAutomaton(getAutomatonName(),this,BuildPhases,debug);
			return automaton;
		}
		else {
			// complete phases!
			automaton.update(BuildPhases);
		}
		return automaton;
	}
}
